"""
Experiments with the Traveling Salesperson Problem (TSP).

A Traveling Salesperson Problem (TSP) is defined as a fully-connected graph
with :attr:`~moptipyapps.tsp.instance.Instance.n_cities` nodes. Each edge in
the graph has a weight, which identifies the distance between the nodes. The
goal is to find the *shortest* tour that visits every single node in the graph
exactly once and then returns back to its starting node. Then nodes are
usually called cities. In this file, we present methods for loading instances
of the TSP as distance matrices `A`. In other words, the value at `A[i, j]`
identifies the travel distance from `i` to `j`. Such instance data can be
loaded via class :mod:`~moptipyapps.tsp.instance`.

In this package, we provide the following tools:

- :mod:`~moptipyapps.tsp.instance` allows you to load instance data in the
  TSPLib format and it provides several instances from TSPLib as resources.
- :mod:`~moptipyapps.tsp.known_optima` provides known optimal solutions for
  some of the TSPLib instances. These should mainly be used for testing
  purposes.
- :mod:`~moptipyapps.tsp.tour_length` is an objective function that can
  efficiently computed the length of a tour in path representation.
- :mod:`~moptipyapps.tsp.tsplib` is just a dummy package holding the actual
  TSPLib data resources.

Important initial work on this code has been contributed by Mr. Tianyu LIANG
(梁天宇), <liangty@stu.hfuu.edu.cn> a Master's student at the Institute of
Applied Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School
of Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei
University (合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the
supervision of Prof. Dr. Thomas Weise (汤卫思教授).
"""



"""
An objective function indirectly minimizing the number of bins in packings.

This objective function first computes the number of bins used. Let's call it
`n_bins`. We know the total number of items,
:attr:`~moptipyapps.binpacking2d.instance.Instance.n_items`, as
well (because this is also the number of rows in the packing).
Now we return `(n_items * (n_bins - 1)) + min_items`,
where `min_items` is the number of items in the bin with the fewest items.

This is similar to :mod:`~moptipyapps.binpacking2d.objectives.\
bin_count_and_last_small`, but instead of focussing on the very last bin, it
uses the minimum element count over all bins. It has the same lower- and upper
bound, though.
"""
from typing import Final

import numba  # type: ignore
import numpy as np

from moptipyapps.binpacking2d.instance import Instance
from moptipyapps.binpacking2d.objectives.bin_count_and_last_empty import (
    BinCountAndLastEmpty,
)
from moptipyapps.binpacking2d.packing import IDX_BIN


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def bin_count_and_empty(y: np.ndarray, temp: np.ndarray) -> int:
    """
    Get the number of bins and number of elements in the emptiest one.

    We compute the total number of bins minus 1 and multiply it with the
    number of items. We then add the smallest number of elements in any
    bin.

    :param y: the packing
    :param temp: the temporary array
    :return: the objective value

    >>> tempa = np.empty(10, int)
    >>> bin_count_and_empty(np.array([[1, 1, 10, 10, 20, 20],
    ...                               [1, 1, 30, 30, 40, 40],
    ...                               [1, 1, 20, 20, 30, 30]], int), tempa)
    3
    >>> bin_count_and_empty(np.array([[1, 1, 10, 10, 20, 20],
    ...                               [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], int), tempa)
    4
    >>> bin_count_and_empty(np.array([[1, 2, 10, 10, 20, 20],  # bin 2!
    ...                               [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], int), tempa)
    4
    >>> bin_count_and_empty(np.array([[1, 3, 10, 10, 20, 20],  # bin 3!
    ...                               [1, 2, 30, 30, 40, 40],  # bin 2!
    ...                               [1, 1, 20, 20, 30, 30]], int), tempa)
    7
    """
    total_bins: int = -1  # the current idea of what the last bin is
    temp.fill(0)  # empty all temporary values
    n_items: Final[int] = len(y)  # the number of rows in the matrix

    for i in range(n_items):  # iterate over all packed items
        bin_idx: int = int(y[i, IDX_BIN]) - 1  # get the bin index of the item
        temp[bin_idx] += 1  # increase number of items
        if bin_idx > total_bins:  # it's a new biggest bin = new last bin?
            total_bins = bin_idx  # and we remember it
    return (n_items * total_bins) + temp[0:total_bins + 1].min()


class BinCountAndEmpty(BinCountAndLastEmpty):
    """Get the number of bins and number of elements in the emptiest one."""

    def __init__(self, instance: Instance) -> None:
        """
        Initialize the objective function.

        :param instance: the instance to load the bounds from
        """
        super().__init__(instance)
        #: the internal temporary array
        self.__temp: Final[np.ndarray] = np.empty(
            instance.n_items, instance.dtype)

    def evaluate(self, x) -> int:
        """
        Evaluate the objective function.

        :param x: the solution
        :return: the bin size and smallest-bin-area factor
        """
        return bin_count_and_empty(x, self.__temp)

    def __str__(self) -> str:
        """
        Get the name of the bins objective function.

        :return: `binCountAndEmpty`
        :retval "binCountAndEmpty": always
        """
        return "binCountAndEmpty"



"""
An extension of the permutation space for one-dimensional ordering.

The main difference is how the `to_str` result is noted, namely that it
contains the mapping of tags to locations.

>>> def _dist(a, b):
...     return abs(a - b)
>>> def _tags(a):
...     return f"t{a}"
>>> the_instance = Instance.from_sequence_and_distance(
...     [1, 2, 3, 3, 2, 3], _tags, _dist)
>>> the_space = OrderingSpace(the_instance)
>>> the_str = the_space.to_str(np.array([0, 2, 1]))
>>> the_str.splitlines()
['0;2;1', '', 'indexZeroBased;suggestedXin01;tag0', '0;0.25;t1', \
'2;0.75;t2', '2;0.75;t2', '1;0.5;t3', '1;0.5;t3', '1;0.5;t3']
>>> print(the_space.from_str(the_str))
[0 2 1]
"""

from io import StringIO
from typing import Final

import numpy as np
from moptipy.spaces.permutations import Permutations
from moptipy.utils.logger import CSV_SEPARATOR
from moptipy.utils.strings import float_to_str
from moptipy.utils.types import type_error

from moptipyapps.order1d.instance import Instance


class OrderingSpace(Permutations):
    """A space for one-dimensional orderings."""

    def __init__(self, instance: Instance) -> None:
        """
        Create an ordering space from an instance.

        :param instance: the instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        super().__init__(range(len(instance)))
        #: the instance
        self.instance: Final[Instance] = instance

    def to_str(self, x: np.ndarray) -> str:
        """
        Convert a solution to a string.

        :param x: the permutation
        :return: the string
        """
        tags: Final[tuple[tuple[tuple[str, ...], int], ...]] \
            = self.instance.tags

        total: Final[int] = len(x) + 1

        with StringIO() as sio:
            sio.writelines(super().to_str(x))
            sio.write("\n\n")

            row: list[str] = ["indexZeroBased", "suggestedXin01"]
            row.extend(f"tag{i}" for i in range(len(tags[0][0])))
            sio.write(CSV_SEPARATOR.join(row))
            sio.write("\n")

            for tag, i in tags:
                row.clear()
                row.append(str(x[i]))
                row.append(float_to_str((x[i] + 1) / total))
                row.extend(tag)
                sio.write(CSV_SEPARATOR.join(row))
                sio.write("\n")
            return sio.getvalue()

    def from_str(self, text: str) -> np.ndarray:
        """
        Get the string version from the given text.

        :param text: the text
        :return: the string
        """
        text = text.lstrip()
        idx: int = text.find("\n")
        if idx > 0:
            text = text[:idx]
        return super().from_str(text.rstrip())



"""
An instance of the Traveling Salesperson Problem (TSP) as distance matrix.

An instance of the Traveling Salesperson Problem (TSP) is defined as a
fully-connected graph with :attr:`Instance.n_cities` nodes. Each edge in the
graph has a weight, which identifies the distance between the nodes. The goal
is to find the *shortest* tour that visits every single node in the graph
exactly once and then returns back to its starting node. Then nodes are
usually called cities. In this file, we present methods for loading instances
of the TSP as distance matrices `A`. In other words, the value at `A[i, j]`
identifies the travel distance from `i` to `j`.

We can load files in a subset of the TSPLib format [1, 2] and also include the
instances of TSPLib with no more than 2000 cities. We load instances as full
distance matrices, which makes writing algorithms to solve them easier but
handling instances with more than 2000 cities problematic due to the high
memory consumption. Of course, you could still load them, but it would be
ill-advised to do so on a normal PC (at the time of this writing).

The TSPLib contains many instances of the symmetric TSP, where the distance
`A[i, j]` from city `i` to city `j` is the same as the distance `A[j, i]` from
`j` to `i`. Here, we provide 110 of them as resource. The cities of some of
these instances are points in the Euclidean plain and the distances are the
(approximate) Euclidean distances. Others use geographical coordinates
(longitude/latitude), and yet others are provides as distances matrices
directly. We also provide 19 of the asymmetric TSPLib instances, where the
distance `A[i, j]` from `i` to `j` may be different from the distance
`A[j, i]` from `j` to `i`.

You can obtain lists of either all, only the symmetric, or only the asymmetric
instance resources via

>>> print(Instance.list_resources()[0:10])  # get all instances (1..10)
('a280', 'ali535', 'att48', 'att532', 'bayg29', 'bays29', 'berlin52', \
'bier127', 'br17', 'brazil58')

>>> print(Instance.list_resources(asymmetric=False)[0:10])  # only symmetric
('a280', 'ali535', 'att48', 'att532', 'bayg29', 'bays29', 'berlin52', \
'bier127', 'brazil58', 'brg180')

>>> print(Instance.list_resources(symmetric=False)[0:10])  # only asymmetric
('br17', 'ft53', 'ft70', 'ftv170', 'ftv33', 'ftv35', 'ftv38', 'ftv44', \
'ftv47', 'ftv55')

You can load any of these instances from the resources via
:meth:`Instance.from_resource` as follows:

>>> inst = Instance.from_resource("a280")
>>> print(inst.n_cities)
280

If you want to read an instance from an external TSPLib file, you can use
:meth:`Instance.from_file`. Be aware that not the whole TSPLib standard is
supported right now, but only a reasonably large subset.

Every TSP instance automatically provides a lower bound
:attr:`Instance.tour_length_lower_bound` and an upper bound
:attr:`Instance.tour_length_upper_bound` of the lengths of valid tours.
For the TSPLib instances, the globally optimal solutions and their tour
lengths are known, so we can use them as lower bounds directly. Otherwise,
we currently use a very crude approximation: We assume that, for each city
`i`, the next city `j` to be visited would be the nearest neighbor of `i`.
Of course, in reality, such a tour may not be feasible, as it could contain
disjoint sets of loops. But no tour can be shorter than this.
As upper bound, we do the same but assume that `j` would be the cities
farthest away from `i`.

>>> print(inst.tour_length_lower_bound)
2579
>>> print(inst.tour_length_upper_bound)
65406

It should be noted that all TSPLib instances by definition have integer
distances. This means that there are never floating point distances and, for
example, Euclidean distances are rounded and are only approximate Euclidean.
Then again, since floating point numbers could also only represent things such
as `sqrt(2)` approximately, using integers instead of floating point numbers
does not really change anything - distances would be approximately Euclidean
or approximately geographical either way.

TSPLib also provides some known optimal solutions in path representation,
i.e., as permutations of the numbers `0` to `n_cities-1`. The optimal
solutions corresponding to the instances provides as resources can be obtained
via :mod:`~moptipyapps.tsp.known_optima`.

The original data of TSPLib can be found at
<http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/>. Before doing
anything with these data directly, you should make sure to read the FAQ
<http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/TSPFAQ.html> and the
documentation
<http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/tsp95.pdf>.

Important initial work on this code has been contributed by Mr. Tianyu LIANG
(梁天宇), <liangty@stu.hfuu.edu.cn> a Master's student at the Institute of
Applied Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School
of Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei
University (合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the
supervision of Prof. Dr. Thomas Weise (汤卫思教授).

1. Gerhard Reinelt. TSPLIB - A Traveling Salesman Problem Library.
   *ORSA Journal on Computing* 3(4):376-384. November 1991.
   https://doi.org/10.1287/ijoc.3.4.376.
   http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/
2. Gerhard Reinelt. *TSPLIB95.* Heidelberg, Germany: Universität
   Heidelberg, Institut für Angewandte Mathematik. 1995.
   http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/tsp95.pdf
3. David Lee Applegate, Robert E. Bixby, Vašek Chvátal, and William John Cook.
   *The Traveling Salesman Problem: A Computational Study.* Second Edition,
   2007. Princeton, NJ, USA: Princeton University Press. Volume 17 of
   Princeton Series in Applied Mathematics. ISBN: 0-691-12993-2.
4. Gregory Z. Gutin and Abraham P. Punnen. *The Traveling Salesman Problem and
   its Variations.* 2002. Kluwer Academic Publishers. Volume 12 of
   Combinatorial Optimization. https://doi.org/10.1007/b101971.
5. Pedro Larrañaga, Cindy M. H. Kuijpers, Roberto H. Murga, Iñaki Inza, and
   Sejla Dizdarevic. Genetic Algorithms for the Travelling Salesman Problem: A
   Review of Representations and Operators. *Artificial Intelligence Review*
   13(2):129-170. April 1999. https://doi.org/10.1023/A:1006529012972.
6. Eugene Leighton Lawler, Jan Karel Lenstra, Alexander Hendrik George Rinnooy
   Kan, and David B. Shmoys. *The Traveling Salesman Problem: A Guided Tour of
   Combinatorial Optimization.* September 1985. Chichester, West Sussex, UK:
   Wiley Interscience. In Estimation, Simulation, and
   Control - Wiley-Interscience Series in Discrete Mathematics and
   Optimization. ISBN: 0-471-90413-9
7. Tianyu Liang, Zhize Wu, Jörg Lässig, Daan van den Berg, and Thomas Weise.
   Solving the Traveling Salesperson Problem using Frequency Fitness
   Assignment. In *Proceedings of the IEEE Symposium on Foundations of
   Computational Intelligence (IEEE FOCI'22),* part of the *IEEE Symposium
   Series on Computational Intelligence (SSCI'22),* December 4-7, 2022,
   Singapore. Pages 360-367. IEEE.
   https://doi.org/10.1109/SSCI51031.2022.10022296.
8. Thomas Weise, Raymond Chiong, Ke Tang, Jörg Lässig, Shigeyoshi Tsutsui,
   Wenxiang Chen, Zbigniew Michalewicz, and Xin Yao. Benchmarking Optimization
   Algorithms: An Open Source Framework for the Traveling Salesman Problem.
   *IEEE Computational Intelligence Magazine.* 9(3):40-52. August 2014.
   https://doi.org/10.1109/MCI.2014.2326101.
"""

from math import acos, cos, isfinite, sqrt
from typing import Any, Callable, Final, TextIO, cast

import moptipy.utils.nputils as npu
import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.path import Path
from moptipy.utils.strings import sanitize_name
from moptipy.utils.types import check_int_range, check_to_int_range, type_error

from moptipyapps.tsp.tsplib import open_resource_stream

#: the known optimal tour lengths and lower bounds of TSP Lib
_LOWER_BOUNDS: dict[str, int] = {
    "a280": 2579, "ali535": 202339, "att48": 10628, "att532": 27686,
    "bayg29": 1610, "bays29": 2020, "berlin52": 7542, "bier127": 118282,
    "br17": 39, "brazil58": 25395, "brd14051": 469385, "brg180": 1950,
    "burma14": 3323, "ch130": 6110, "ch150": 6528, "d1291": 50801,
    "d15112": 1573084, "d1655": 62128, "d18512": 645238, "d198": 15780,
    "d2103": 80450, "d493": 35002, "d657": 48912, "dantzig42": 699,
    "dsj1000": 18660188, "eil101": 629, "eil51": 426, "eil76": 538,
    "fl1400": 20127, "fl1577": 22249, "fl3795": 28772, "fl417": 11861,
    "fnl4461": 182566, "fri26": 937, "ft53": 6905, "ft70": 38673,
    "ftv100": 1788, "ftv110": 1958, "ftv120": 2166, "ftv130": 2307,
    "ftv140": 2420, "ftv150": 2611, "ftv160": 2683, "ftv170": 2755,
    "ftv33": 1286, "ftv35": 1473, "ftv38": 1530, "ftv44": 1613, "ftv47": 1776,
    "ftv55": 1608, "ftv64": 1839, "ftv70": 1950, "ftv90": 1579,
    "gil262": 2378, "gr120": 6942, "gr137": 69853, "gr17": 2085,
    "gr202": 40160, "gr21": 2707, "gr229": 134602, "gr24": 1272,
    "gr431": 171414, "gr48": 5046, "gr666": 294358, "gr96": 55209,
    "hk48": 11461, "kro124p": 36230, "kroA100": 21282, "kroA150": 26524,
    "kroA200": 29368, "kroB100": 22141, "kroB150": 26130, "kroB200": 29437,
    "kroC100": 20749, "kroD100": 21294, "kroE100": 22068, "lin105": 14379,
    "lin318": 42029, "nrw1379": 56638, "p43": 5620, "p654": 34643,
    "pa561": 2763, "pcb1173": 56892, "pcb3038": 137694, "pcb442": 50778,
    "pla33810": 66048945, "pla7397": 23260728, "pla85900": 142382641,
    "pr1002": 259045, "pr107": 44303, "pr124": 59030, "pr136": 96772,
    "pr144": 58537, "pr152": 73682, "pr226": 80369, "pr2392": 378032,
    "pr264": 49135, "pr299": 48191, "pr439": 107217, "pr76": 108159,
    "rat195": 2323, "rat575": 6773, "rat783": 8806, "rat99": 1211,
    "rbg323": 1326, "rbg358": 1163, "rbg403": 2465, "rbg443": 2720,
    "rd100": 7910, "rd400": 15281, "rl11849": 923288, "rl1304": 252948,
    "rl1323": 270199, "rl1889": 316536, "rl5915": 565530, "rl5934": 556045,
    "ry48p": 14422, "si1032": 92650, "si175": 21407, "si535": 48450,
    "st70": 675, "swiss42": 1273, "ts225": 126643, "tsp225": 3916,
    "u1060": 224094, "u1432": 152970, "u159": 42080, "u1817": 57201,
    "u2152": 64253, "u2319": 234256, "u574": 36905, "u724": 41910,
    "ulysses16": 6859, "ulysses22": 7013, "usa13509": 19982859,
    "vm1084": 239297, "vm1748": 336556}

#: the TSPLib instances
_SYMMETRIC_INSTANCES: Final[tuple[str, ...]] = (
    "a280", "ali535", "att48", "att532", "bayg29", "bays29", "berlin52",
    "bier127", "brazil58", "brg180", "burma14", "ch130", "ch150", "d1291",
    "d1655", "d198", "d493", "d657", "dantzig42", "eil101", "eil51", "eil76",
    "fl1400", "fl1577", "fl417", "fri26", "gil262", "gr120", "gr137", "gr17",
    "gr202", "gr21", "gr229", "gr24", "gr431", "gr48", "gr666", "gr96",
    "hk48", "kroA100", "kroA150", "kroA200", "kroB100", "kroB150", "kroB200",
    "kroC100", "kroD100", "kroE100", "lin105", "lin318", "nrw1379", "p654",
    "pcb1173", "pcb442", "pr1002", "pr107", "pr124", "pr136", "pr144",
    "pr152", "pr226", "pr264", "pr299", "pr439", "pr76", "rat195", "rat575",
    "rat783", "rat99", "rd100", "rd400", "rl1304", "rl1323", "rl1889",
    "si1032", "si175", "si535", "st70", "swiss42", "ts225", "tsp225", "u1060",
    "u1432", "u159", "u1817", "u574", "u724", "ulysses16", "ulysses22",
    "vm1084", "vm1748")

#: The set of asymmetric instances
_ASYMMETRIC_INSTANCES: Final[tuple[str, ...]] = (
    "br17", "ft53", "ft70", "ftv170", "ftv33", "ftv35", "ftv38", "ftv44",
    "ftv47", "ftv55", "ftv64", "ftv70", "kro124p", "p43", "rbg323", "rbg358",
    "rbg403", "rbg443", "ry48p")

#: The set of all TSP instances
_INSTANCES: Final[tuple[str, ...]] = tuple(sorted(
    _SYMMETRIC_INSTANCES + _ASYMMETRIC_INSTANCES))

#: the set of asymmetric resources
_ASYMMETRIC_RESOURCES: Final[set[str]] = set(_ASYMMETRIC_INSTANCES)

#: the problem is a symmetric tsp
__TYPE_SYMMETRIC_TSP: Final[str] = "TSP"
#: the problem is an asymmetric tsp
__TYPE_ASYMMETRIC_TSP: Final[str] = "ATSP"
#: the permitted types
__TYPES: Final[set[str]] = {__TYPE_SYMMETRIC_TSP, __TYPE_ASYMMETRIC_TSP}
#: the name start
__KEY_NAME: Final[str] = "NAME"
#: the type start
__KEY_TYPE: Final[str] = "TYPE"
#: the dimension start
__KEY_DIMENSION: Final[str] = "DIMENSION"
#: the comment start
__KEY_COMMENT: Final[str] = "COMMENT"
#: EUC_2D coordinates
__EWT_EUC_2D: Final[str] = "EUC_2D"
#: geographical coordinates
__EWT_GEO: Final[str] = "GEO"
#: ATT coordinates
__EWT_ATT: Final[str] = "ATT"
#: ceiling 2D coordinates
__EWT_CEIL2D: Final[str] = "CEIL_2D"
#: the explicit edge weight type
__EWT_EXPLICIT: Final[str] = "EXPLICIT"
#: the edge weight type start
__KEY_EDGE_WEIGHT_TYPE: Final[str] = "EDGE_WEIGHT_TYPE"
#: the permitted edge weight types
__EDGE_WEIGHT_TYPES: Final[set[str]] = {
    __EWT_EUC_2D, __EWT_GEO, __EWT_ATT, __EWT_CEIL2D, __EWT_EXPLICIT}
#: the edge weight format "function"
__EWF_FUNCTION: Final[str] = "FUNCTION"
#: the full matrix edge weight format
__EWF_FULL_MATRIX: Final[str] = "FULL_MATRIX"
#: the upper row edge weight format
__EWF_UPPER_ROW: Final[str] = "UPPER_ROW"
#: the lower diagonal row
__EWF_LOWER_DIAG_ROW: Final[str] = "LOWER_DIAG_ROW"
#: the upper diagonal row
__EWF_UPPER_DIAG_ROW: Final[str] = "UPPER_DIAG_ROW"
#: the edge weight format start
__KEY_EDGE_WEIGHT_FORMAT: Final[str] = "EDGE_WEIGHT_FORMAT"
#: the permitted edge weight formats
__EDGE_WEIGHT_FORMATS: Final[set[str]] = {
    __EWF_FUNCTION, __EWF_FULL_MATRIX, __EWF_UPPER_ROW,
    __EWF_LOWER_DIAG_ROW, __EWF_UPPER_DIAG_ROW}
#: the start of the node coord type
__KEY_NODE_COORD_TYPE: Final[str] = "NODE_COORD_TYPE"
#: 2d coordinates
__NODE_COORD_TYPE_2D: Final[str] = "TWOD_COORDS"
#: no coordinates
__NODE_COORD_TYPE_NONE: Final[str] = "NO_COORDS"
#: the permitted node coordinate types
__NODE_COORD_TYPES: Final[set[str]] = {
    __NODE_COORD_TYPE_2D, __NODE_COORD_TYPE_NONE}
#: the node coordinate section starts
__START_NODE_COORD_SECTION: Final[str] = "NODE_COORD_SECTION"
#: start the edge weight section
__START_EDGE_WEIGHT_SECTION: Final[str] = "EDGE_WEIGHT_SECTION"
#: the end of the file
__EOF: Final[str] = "EOF"
#: the fixed edges section
__FIXED_EDGES: Final[str] = "FIXED_EDGES_SECTION"


def __line_to_nums(line: str,
                   collector: Callable[[int | float], Any]) -> None:
    """
    Split a line to a list of integers.

    :param line: the line string
    :param collector: the collector
    """
    if not isinstance(line, str):
        raise type_error(line, "line", str)
    idx: int = 0
    line = line.strip()
    str_len: Final[int] = len(line)

    while idx < str_len:
        while (idx < str_len) and (line[idx] == " "):
            idx += 1
        if idx >= str_len:
            return

        next_space = line.find(" ", idx)
        if next_space < 0:
            next_space = str_len

        part: str = line[idx:next_space]
        if ("." in part) or ("E" in part) or ("e" in part):
            f: float = float(part)
            if not isfinite(f):
                raise ValueError(
                    f"{part!r} translates to non-finite float {f}.")
            collector(f)
        else:
            collector(check_to_int_range(
                part, "line fragment", -1_000_000_000_000, 1_000_000_000_000))
        idx = next_space


def __nint(v: int | float) -> int:
    """
    Get the nearest integer in the TSPLIB format.

    :param v: the value
    :return: the integer
    """
    if isinstance(v, int):
        return v
    if isinstance(v, float):
        return int(0.5 + v)
    raise type_error(v, "value", (int, float))


def __dist_2deuc(a: list[int | float], b: list[int | float]) -> int:
    """
    Compute the two-dimensional Euclidean distance function.

    :param a: the first point
    :param b: the second point
    :return: the distance
    """
    return __nint(sqrt(((a[0] - b[0]) ** 2) + ((a[1] - b[1]) ** 2)))


def __matrix_from_points(
        n_cities: int, coord_dim: int, stream: TextIO,
        dist_func: Callable[[list[int | float], list[int | float]], int]) \
        -> np.ndarray:
    """
    Load a distance matrix from 2D coordinates.

    :param n_cities: the dimension
    :param coord_dim: the coordinate dimension
    :param stream: the stream
    :param dist_func: the distance function
    :return: the matrix
    """
    index: int = 0
    coordinates: Final[list[list[int | float]]] = []
    row: Final[list[int | float]] = []
    for the_line in stream:
        if not isinstance(the_line, str):
            raise type_error(the_line, "line", str)
        line = the_line.strip()
        if len(line) <= 0:
            continue
        if line == __EOF:
            break
        __line_to_nums(line, row.append)
        index += 1
        if (len(row) != (coord_dim + 1)) or (not isinstance(row[0], int)) \
                or (row[0] != index):
            raise ValueError(f"invalid row {line!r} at index {index}, "
                             f"gives values {row}.")
        coordinates.append(row[1:])
        row.clear()
    if index != n_cities:
        raise ValueError(f"only found {index} rows, but expected {n_cities}")

    # now construct the matrix
    matrix: Final[np.ndarray] = np.zeros((n_cities, n_cities),
                                         npu.DEFAULT_INT)
    for i in range(n_cities):
        a = coordinates[i]
        for j in range(i):
            b = coordinates[j]
            dist = dist_func(a, b)
            if not isinstance(dist, int):
                raise type_error(dist, f"distance[{i},{j}]", int)
            if dist < 0:
                raise ValueError(
                    f"Cannot have node distance {dist}: ({a}) at index="
                    f"{i + 1} and ({b} at index={j + 1}.")
            matrix[i, j] = dist
            matrix[j, i] = dist
    return matrix


def __coord_to_rad(x: int | float) -> float:
    """
    Convert coordinates to longitude/latitude.

    :param x: the coordinate
    """
    degrees: int = int(x)
    return (3.141592 * (degrees + (5.0 * (x - degrees)) / 3.0)) / 180.0


def __dist_loglat(a: list[int | float], b: list[int | float]) -> int:
    """
    Convert a longitude-latitude pair to a distance.

    :param a: the first point
    :param b: the second point
    :return: the distance
    """
    lat1: float = __coord_to_rad(a[0])
    long1: float = __coord_to_rad(a[1])
    lat2: float = __coord_to_rad(b[0])
    long2: float = __coord_to_rad(b[1])
    q1: Final[float] = cos(long1 - long2)
    q2: Final[float] = cos(lat1 - lat2)
    q3: Final[float] = cos(lat1 + lat2)
    return int(6378.388 * acos(0.5 * (
        (1.0 + q1) * q2 - (1.0 - q1) * q3)) + 1.0)


def __dist_att(a: list[int | float], b: list[int | float]) -> int:
    """
    Compute the ATT Pseudo-Euclidean distance function.

    :param a: the first point
    :param b: the second point
    :return: the distance
    """
    xd: Final[int | float] = a[0] - b[0]
    yd: Final[int | float] = a[1] - b[1]
    rij: Final[float] = sqrt((xd * xd + yd * yd) / 10.0)
    tij: Final[int] = __nint(rij)
    return (tij + 1) if tij < rij else tij


def __dist_2dceil(a: list[int | float], b: list[int | float]) -> int:
    """
    Compute the ceiling of the two-dimensional Euclidean distance function.

    :param a: the first point
    :param b: the second point
    :return: the distance
    """
    dist: Final[float] = sqrt(((a[0] - b[0]) ** 2) + ((a[1] - b[1]) ** 2))
    disti: Final[int] = int(dist)
    return disti if dist == disti else (disti + 1)


def _matrix_from_node_coord_section(
        n_cities: int | None, edge_weight_type: str | None,
        node_coord_type: str | None, stream: TextIO) -> np.ndarray:
    """
    Get a data matrix from a node coordinate section.

    :param n_cities: the dimension
    :param edge_weight_type: the edge weight type
    :param node_coord_type: the node coordinate type
    :param stream: the node coordinate stream
    :return: the data matrix
    """
    check_int_range(n_cities, "n_cities", 2, 1_000_000_000_000)
    if (node_coord_type is not None) and \
            (not isinstance(node_coord_type, str)):
        raise type_error(node_coord_type, "node_coord_type", str)
    if not isinstance(edge_weight_type, str):
        raise type_error(edge_weight_type, "edge_weight_type", str)

    dist_fun = None
    coord_dim: int | None = None
    if (edge_weight_type == __EWT_EUC_2D) \
            and (node_coord_type in (None, __NODE_COORD_TYPE_2D)):
        coord_dim = 2
        dist_fun = __dist_2deuc
    elif (edge_weight_type == __EWT_GEO) \
            and (node_coord_type in (None, __NODE_COORD_TYPE_2D)):
        coord_dim = 2
        dist_fun = __dist_loglat
    elif (edge_weight_type == __EWT_ATT) \
            and (node_coord_type in (None, __NODE_COORD_TYPE_2D)):
        coord_dim = 2
        dist_fun = __dist_att
    elif (edge_weight_type == __EWT_CEIL2D) \
            and (node_coord_type in (None, __NODE_COORD_TYPE_2D)):
        coord_dim = 2
        dist_fun = __dist_2dceil

    if (coord_dim is not None) and (dist_fun is not None):
        return __matrix_from_points(n_cities, coord_dim, stream, dist_fun)

    raise ValueError(f"invalid combination of {__KEY_EDGE_WEIGHT_TYPE} "
                     f"and {__KEY_NODE_COORD_TYPE}")


def __read_n_ints(n: int, stream: TextIO) -> list[int]:
    """
    Read exactly `n` integers from a stream.

    :param n: the number of integers to read
    :param stream: the stream to read from
    :return: the list of integers
    """
    res: list[int] = []

    def __append(i: int | float, fwd=res.append) -> None:
        if isinstance(i, int):
            fwd(i)
        else:
            i2 = int(i)
            if i2 != i:
                raise ValueError(f"{i} is not integer")
            fwd(i2)

    for line in stream:
        __line_to_nums(line, __append)
        if len(res) == n:
            break

    if len(res) != n:
        raise ValueError(f"expected {n} integers, got {len(res)}.")
    return res


def _matrix_from_edge_weights(
        n_cities: int | None, edge_weight_type: str | None,
        edge_weight_format: str | None, stream: TextIO) -> np.ndarray:
    """
    Get a data matrix from a edge weights section.

    :param n_cities: the dimension
    :param edge_weight_type: the edge weight type
    :param edge_weight_format: the edge weight format
    :param stream: the node coordinate stream
    :return: the data matrix
    """
    check_int_range(n_cities, "n_cities", 2, 1_000_000_000_000)
    if not isinstance(edge_weight_type, str):
        raise type_error(edge_weight_type, "node_coord_type", str)
    if not isinstance(edge_weight_type, str):
        raise type_error(edge_weight_type, "edge_weight_type", str)
    if edge_weight_type == __EWT_EXPLICIT:
        if edge_weight_format == __EWF_FULL_MATRIX:
            res = np.array(__read_n_ints(n_cities * n_cities, stream),
                           dtype=npu.DEFAULT_INT).reshape(
                (n_cities, n_cities))
            np.fill_diagonal(res, 0)
            return res
        if edge_weight_format == __EWF_UPPER_ROW:
            ints = __read_n_ints((n_cities * (n_cities - 1)) // 2, stream)
            res = np.zeros((n_cities, n_cities), dtype=npu.DEFAULT_INT)
            i: int = 1
            j: int = 0
            for v in ints:
                res[j, i] = res[i, j] = v
                i = i + 1
                if i >= n_cities:
                    j = j + 1
                    i = j + 1
            return res
        if edge_weight_format == __EWF_LOWER_DIAG_ROW:
            ints = __read_n_ints(
                n_cities + ((n_cities * (n_cities - 1)) // 2), stream)
            res = np.zeros((n_cities, n_cities), dtype=npu.DEFAULT_INT)
            i = 0
            j = 0
            for v in ints:
                if i != j:
                    res[j, i] = res[i, j] = v
                i = i + 1
                if i > j:
                    j = j + 1
                    i = 0
            return res
        if edge_weight_format == __EWF_UPPER_DIAG_ROW:
            ints = __read_n_ints(
                n_cities + ((n_cities * (n_cities - 1)) // 2), stream)
            res = np.zeros((n_cities, n_cities), dtype=npu.DEFAULT_INT)
            i = 0
            j = 0
            for v in ints:
                if i != j:
                    res[j, i] = res[i, j] = v
                i = i + 1
                if i >= n_cities:
                    j = j + 1
                    i = 0
            return res
    raise ValueError(
        f"unsupported combination of {__KEY_EDGE_WEIGHT_TYPE}="
        f"{edge_weight_type!r} and {__KEY_EDGE_WEIGHT_FORMAT}="
        f"{edge_weight_format!r}")


def _from_stream(
        stream: TextIO,
        lower_bound_getter: Callable[[str], int] | None =
        _LOWER_BOUNDS.__getitem__) -> "Instance":
    """
    Read a TSP Lib instance from a TSP-lib formatted stream.

    :param stream: the text stream
    :param lower_bound_getter: a function returning a lower bound for an
        instance name, or `None` to use a simple lower bound approximation
    :return: the instance
    """
    if (lower_bound_getter is not None) \
            and (not callable(lower_bound_getter)):
        raise type_error(
            lower_bound_getter, "lower_bound_getter", None, call=True)

    the_name: str | None = None
    the_type: str | None = None
    the_n_cities: int | None = None
    the_ewt: str | None = None
    the_ewf: str | None = None
    the_nct: str | None = None
    the_matrix: np.ndarray | None = None

    for the_line in stream:
        if not isinstance(the_line, str):
            raise type_error(the_line, "line", str)
        line = the_line.strip()
        if len(line) <= 0:
            continue

        sep_idx: int = line.find(":")
        if sep_idx > 0:
            key: str = line[:sep_idx].strip()
            value: str = line[sep_idx + 1:].strip()
            if len(value) <= 0:
                raise ValueError(f"{line!r} has empty value "
                                 f"{value!r} for key {key!r}.")

            if key == __KEY_NAME:
                if the_name is not None:
                    raise ValueError(
                        f"{line!r} cannot come at this position,"
                        f" already got {__KEY_NAME}={the_name!r}.")
                if value.endswith(".tsp"):
                    value = value[0:-4]
                the_name = value
                continue

            if key == __KEY_TYPE:
                if the_type is not None:
                    raise ValueError(
                        f"{line!r} cannot come at this position, already "
                        f"got {__KEY_TYPE}={the_type!r}.")
                the_type = __TYPE_SYMMETRIC_TSP \
                    if value == "TSP (M.~Hofmeister)" else value
                if the_type not in __TYPES:
                    raise ValueError(
                        f"only {__TYPES!r} are permitted as {__KEY_TYPE}, "
                        f"but got {the_type!r} from {line!r}.")
                continue

            if key == __KEY_DIMENSION:
                if the_n_cities is not None:
                    raise ValueError(
                        f"{line!r} cannot come at this position,"
                        f" already got {__KEY_DIMENSION}={the_n_cities}.")
                the_n_cities = check_to_int_range(value, "dimension",
                                                  2, 1_000_000_000)

            if key == __KEY_EDGE_WEIGHT_TYPE:
                if the_ewt is not None:
                    raise ValueError(
                        f"{line!r} cannot come at this position, already "
                        f"got {__KEY_EDGE_WEIGHT_TYPE}={the_ewt!r}.")
                the_ewt = value
                if the_ewt not in __EDGE_WEIGHT_TYPES:
                    raise ValueError(
                        f"only {__EDGE_WEIGHT_TYPES!r} are permitted as "
                        f"{__KEY_EDGE_WEIGHT_TYPE}, but got {the_ewt!r} "
                        f"in {line!r}")
                continue

            if key == __KEY_EDGE_WEIGHT_FORMAT:
                if the_ewf is not None:
                    raise ValueError(
                        f"{line!r} cannot come at this position, already "
                        f"got {__KEY_EDGE_WEIGHT_FORMAT}={the_ewf!r}.")
                the_ewf = value
                if the_ewf not in __EDGE_WEIGHT_FORMATS:
                    raise ValueError(
                        f"only {__EDGE_WEIGHT_FORMATS!r} are permitted as "
                        f"{__KEY_EDGE_WEIGHT_FORMAT}, but got {the_ewf} "
                        f"in {line!r}")
                continue
            if key == __KEY_NODE_COORD_TYPE:
                if the_nct is not None:
                    raise ValueError(
                        f"{line!r} cannot come at this position, already "
                        f"got {__KEY_NODE_COORD_TYPE}={the_nct!r}.")
                the_nct = value
                if the_nct not in __NODE_COORD_TYPES:
                    raise ValueError(
                        f"only {__NODE_COORD_TYPES!r} are permitted as node "
                        f"{__KEY_NODE_COORD_TYPE}, but got {the_nct!r} "
                        f"in {line!r}")
                continue
        elif line == __START_NODE_COORD_SECTION:
            if the_matrix is not None:
                raise ValueError(
                    f"already got matrix, cannot have {line!r} here!")
            the_matrix = _matrix_from_node_coord_section(
                the_n_cities, the_ewt, the_nct, stream)
            continue
        elif line == __START_EDGE_WEIGHT_SECTION:
            if the_matrix is not None:
                raise ValueError(
                    f"already got matrix, cannot have {line!r} here!")
            the_matrix = _matrix_from_edge_weights(
                the_n_cities, the_ewt, the_ewf, stream)
            continue
        elif line == __EOF:
            break
        elif line == __FIXED_EDGES:
            raise ValueError(f"{__FIXED_EDGES!r} not supported")

    if the_name is None:
        raise ValueError("did not find any name.")
    if the_matrix is None:
        raise ValueError(f"did not find any matrix for {the_name!r}.")

    inst: Final[Instance] = Instance(
        the_name, 0 if (lower_bound_getter is None) else
        lower_bound_getter(the_name), the_matrix)
    if (the_type == __TYPE_SYMMETRIC_TSP) and (not inst.is_symmetric):
        raise ValueError("found asymmetric TSP instance but expected "
                         f"{the_name!r} to be a symmetric one?")

    return inst


def ncities_from_tsplib_name(name: str) -> int:
    """
    Compute the instance scale from the instance name.

    :param name: the instance name
    :return: the instance scale
    """
    if name == "kro124p":
        return 100
    if name == "ry48p":
        return 48
    idx: int = len(name)
    while name[idx - 1] in "0123456789":
        idx -= 1
    scale: Final[int] = int(name[idx:])
    if name.startswith("ftv"):
        return scale + 1
    return scale


class Instance(Component, np.ndarray):
    """An instance of the Traveling Salesperson Problem."""

    #: the name of the instance
    name: str
    #: the number of cities
    n_cities: int
    #: the lower bound of the tour length
    tour_length_lower_bound: int
    #: the upper bound of the tour length
    tour_length_upper_bound: int
    #: is the TSP instance symmetric?
    is_symmetric: bool

    def __new__(cls, name: str, tour_length_lower_bound: int,
                matrix: np.ndarray,
                upper_bound_range_multiplier: int = 1) -> "Instance":
        """
        Create an instance of the Traveling Salesperson Problem.

        :param cls: the class
        :param name: the name of the instance
        :param tour_length_lower_bound: the lower bound of the tour length
        :param matrix: the matrix with the data (will be copied)
        :param upper_bound_range_multiplier: a multiplier for the upper bound
            to fix the data range, by default `1`
        """
        use_name: Final[str] = sanitize_name(name)
        if name != use_name:
            raise ValueError(f"Name {name!r} is not a valid name.")

        check_int_range(tour_length_lower_bound, "tour_length_lower_bound",
                        0, 1_000_000_000_000_000)

        n_cities: int = len(matrix)
        if n_cities <= 1:
            raise ValueError(f"There must be at least two cities in a TSP "
                             f"instance, but we got {n_cities}.")

        use_shape: Final[tuple[int, int]] = (n_cities, n_cities)
        if isinstance(matrix, np.ndarray):
            if matrix.shape != use_shape:
                raise ValueError(
                    f"Unexpected shape {matrix.shape} for {n_cities} cities, "
                    f"expected {use_shape}.")
        else:
            raise type_error(matrix, "matrix", np.ndarray)

        # validate the matrix and compute the upper bound
        upper_bound: int = 0
        lower_bound_2: int = 0
        is_symmetric: bool = True
        for i in range(n_cities):
            farthest_neighbor: int = -1
            nearest_neighbor: int = 9_223_372_036_854_775_807

            for j in range(n_cities):
                dist = int(matrix[i, j])
                if i == j:
                    if dist != 0:
                        raise ValueError(
                            f"if i=j={i}, then dist must be zero "
                            f"but is {dist}.")
                    continue
                if dist > farthest_neighbor:
                    farthest_neighbor = dist
                if dist < nearest_neighbor:
                    nearest_neighbor = dist
                if dist != matrix[j, i]:
                    is_symmetric = False
            if farthest_neighbor <= 0:
                raise ValueError(f"farthest neighbor distance of node {i} is"
                                 f" {farthest_neighbor}?")
            upper_bound = upper_bound + farthest_neighbor
            lower_bound_2 = lower_bound_2 + nearest_neighbor

        tour_length_lower_bound = max(
            tour_length_lower_bound, check_int_range(
                lower_bound_2, "lower_bound_2", 0, 1_000_000_000_000_000))
        check_int_range(upper_bound, "upper_bound",
                        tour_length_lower_bound, 1_000_000_000_000_001)

        # create the object and ensure that the backing integer type is
        # large enough to accommodate all possible tour lengths
        limit: Final[int] = (check_int_range(
            upper_bound_range_multiplier, "upper_bound_range_multiplier", 1)
            * max(upper_bound, n_cities))
        obj: Final[Instance] = super().__new__(
            cls, use_shape, int_range_to_dtype(
                min_value=-limit, max_value=limit))
        np.copyto(obj, matrix, "unsafe")
        for i in range(n_cities):
            for j in range(n_cities):
                if obj[i, j] != matrix[i, j]:
                    raise ValueError(
                        f"error when copying: {obj[i, j]} != {matrix[i, j]}")

        #: the name of the instance
        obj.name = use_name
        #: the number of cities
        obj.n_cities = n_cities
        #: the lower bound of the tour length
        obj.tour_length_lower_bound = tour_length_lower_bound
        #: the upper bound of the tour length
        obj.tour_length_upper_bound = upper_bound
        #: is this instance symmetric?
        obj.is_symmetric = is_symmetric
        return obj

    def __str__(self):
        """
        Get the name of this instance.

        :return: the name of this instance

        >>> str(Instance.from_resource("gr17"))
        'gr17'
        """
        return self.name

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the instance to the given logger.

        :param logger: the logger for the parameters

        >>> from moptipy.utils.logger import InMemoryLogger
        >>> with InMemoryLogger() as l:
        ...     with l.key_values("I") as kv:
        ...         Instance.from_resource("kroE100").log_parameters_to(kv)
        ...     print(repr('@'.join(l.get_log())))
        'BEGIN_I@name: kroE100@class: moptipyapps.tsp.instance.Instance\
@nCities: 100@tourLengthLowerBound: 22068@tourLengthUpperBound: 330799@\
symmetric: T@dtype: i@END_I'
        """
        super().log_parameters_to(logger)
        logger.key_value("nCities", self.n_cities)
        logger.key_value("tourLengthLowerBound", self.tour_length_lower_bound)
        logger.key_value("tourLengthUpperBound", self.tour_length_upper_bound)
        logger.key_value("symmetric", self.is_symmetric)
        logger.key_value(npu.KEY_NUMPY_TYPE, self.dtype.char)

    @staticmethod
    def from_file(
            path: str,
            lower_bound_getter: Callable[[str], int] | None =
            _LOWER_BOUNDS.__getitem__) -> "Instance":
        """
        Read a TSP Lib instance from a TSP-lib formatted file.

        :param path: the path to the file
        :param lower_bound_getter: a function returning a lower bound for an
            instance name, or `None` to use a simple lower bound approximation
        :return: the instance

        >>> from os.path import dirname
        >>> inst = Instance.from_file(dirname(__file__) + "/tsplib/br17.atsp")
        >>> inst.name
        'br17'
        """
        file: Final[Path] = Path.file(path)
        with file.open_for_read() as stream:
            try:
                return _from_stream(
                    cast(TextIO, stream), lower_bound_getter)
            except (TypeError, ValueError) as err:
                raise ValueError(f"error when parsing file {file!r}") from err

    @staticmethod
    def from_resource(name: str) -> "Instance":
        """
        Load a TSP instance from a resource.

        :param name: the name string
        :return: the instance

        >>> Instance.from_resource("br17").n_cities
        17
        """
        if not isinstance(name, str):
            raise type_error(name, "name", str)
        container: Final = Instance.from_resource
        inst_attr: Final[str] = f"__inst_{name}"
        if hasattr(container, inst_attr):  # instance loaded?
            return cast(Instance, getattr(container, inst_attr))

        is_symmetric: Final[bool] = name not in _ASYMMETRIC_INSTANCES
        suffix: Final[str] = ".tsp" if is_symmetric else ".atsp"
        with open_resource_stream(f"{name}{suffix}") as stream:
            inst: Final[Instance] = _from_stream(stream)

        if inst.name != name:
            raise ValueError(f"got {inst.name!r} for instance {name!r}?")
        sc: int = ncities_from_tsplib_name(name)
        if sc != inst.n_cities:
            raise ValueError(f"expected to get n_cities={sc} from name "
                             f"{name!r} but got {inst.n_cities}.")
        if is_symmetric and (not inst.is_symmetric):
            raise ValueError(f"{name!r} should be symmetric but is not?")
        if inst.n_cities <= 1000:
            setattr(container, inst_attr, inst)
        return inst

    @staticmethod
    def list_resources(symmetric: bool = True,
                       asymmetric: bool = True) -> tuple[str, ...]:
        """
        Get a tuple of all the TSP instances available as resource.

        :param symmetric: include the symmetric instances
        :param asymmetric: include the asymmetric instances
        :return: the tuple with the instance names

        >>> a = len(Instance.list_resources(True, True))
        >>> print(a)
        110
        >>> b = len(Instance.list_resources(True, False))
        >>> print(b)
        91
        >>> c = len(Instance.list_resources(False, True))
        >>> print(c)
        19
        >>> print(a == (b + c))
        True
        >>> print(len(Instance.list_resources(False, False)))
        0
        """
        return _INSTANCES if (symmetric and asymmetric) else (
            _SYMMETRIC_INSTANCES if symmetric else (
                _ASYMMETRIC_INSTANCES if asymmetric else ()))



"""
An objective function evaluating a permutation of objects.

The goal is to find an order of objects such that the ranks of their distances
inside the permutation match to the distance ranks inside the instance data.

In other words, we take a permutation and consider the distance between two
elements the differences of their indices in th permutation. So we can
translate a permutation to a distance matrix. Then we transform the distance
matrix into a "distance rank matrix", where for each object (row), we compute
the distance rank of each neighbor, exactly as we do in the problem
:mod:`~moptipyapps.order1d.instance`. Ideally, the resulting distance rank
matrix should be exactly the same as the :mod:`~moptipyapps.order1d.instance`,
which is also such a matrix.

As objective value, we first compute the element-wise absolute difference
between both matrices and element-wise divide it by the element-wise square
of the instance matrix. We then return the arithmetic mean over all elements
in the result matrix. (Actually, using the square is just default setting,
you can use any power you like...)

If this mean value hits `0.0`, we have found an exact mapping of objects to
positions. Normally, we will not be able to reach such a value, though. But
the closer we get to `0`, the better should the representation be.

The objective function is symmetric with respect to the permutation, i.e.,
ordering the objects à la `[0, 1, 2, 3]` is the same as ordering them in
`[3, 2, 1, 0]`.

First, let's construct an instance.
>>> def _dist(a, b):
...     return abs(a - b)
>>> def _tags(a):
...     return f"t{a}"
>>> the_instance = Instance.from_sequence_and_distance(
...     [1, 2, 3, 4, 5], _tags, _dist)

Now let's create the objective function:
>>> the_objective = OneDimensionalDistribution(the_instance)
>>> the_x = np.array([0, 1, 2, 3, 4], int)
>>> the_objective.evaluate(the_x)
0.0

>>> the_x = np.array([1, 0, 2, 3, 4], int)
>>> the_objective.evaluate(the_x)
0.02878505920886873

>>> the_x = np.array([2, 1, 0, 3, 4], int)
>>> the_objective.evaluate(the_x)
0.04803169564121945

>>> the_x = np.array([4, 3, 2, 1, 0], int)
>>> the_objective.evaluate(the_x)
0.0
"""


from math import isfinite
from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import num_to_str
from moptipy.utils.types import type_error
from scipy.stats import rankdata  # type: ignore

from moptipyapps.order1d.instance import Instance
from moptipyapps.shared import SCOPE_INSTANCE


@numba.njit(cache=True, inline="always", fastmath=True, boundscheck=False)
def _get_dist(x: np.ndarray, out: np.ndarray) -> None:
    """
    Fill a distance matrix.

    :param x: the permutation
    :param out: the output matrix

    >>> a = np.array([0, 1, 2, 3], int)
    >>> res = np.empty((len(a), len(a)), int)
    >>> _get_dist(a, res)
    >>> print(res)
    [[0 1 2 3]
     [1 0 1 2]
     [2 1 0 1]
     [3 2 1 0]]
    >>> a = np.array([2, 1, 0, 3], int)
    >>> _get_dist(a, res)
    >>> print(res)
    [[0 1 2 1]
     [1 0 1 2]
     [2 1 0 3]
     [1 2 3 0]]
    """
    lenx: Final[int] = len(x)
    for i1, x1 in enumerate(x):
        out[i1, i1] = 0
        for i2 in range(i1 + 1, lenx):
            x2 = x[i2]
            out[x1, x2] = out[x2, x1] = i2 - i1


class OneDimensionalDistribution(Objective):
    """An objective function for the one-dimensional ordering problem."""

    def __init__(self, instance: Instance, power: float | int = 2) -> None:
        """
        Initialize the one-dimensional distribution objective function.

        :param instance: the one-dimensional ordering problem.
        :param power: the power to be used for the distance rank scaling
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        if not isinstance(power, int | float):
            raise type_error(power, "power", (int, float))
        if not (isfinite(power) and (power > 1)):
            raise ValueError(f"invalid power, must be > 1 but is {power}.")
        #: the power
        self.power: Final[int | float] = power
        #: the instance data
        self.instance: Final[Instance] = instance
        #: the instance data by the power of minus three, to speed up stuff
        self.__inst3: Final[np.ndarray] = instance ** float(-power)
        #: a temporary array
        self.__temp: Final[np.ndarray] = np.empty(
            instance.shape, instance.dtype)

    def evaluate(self, x) -> float:
        """
        Get the difference between the element ordering and the rank goals.

        :param x: the permutation of elements
        :return: the difference between the rank goals and orderings
        """
        temp: Final[np.ndarray] = self.__temp
        _get_dist(x, temp)
        rd: np.ndarray = (2.0 * rankdata(temp, axis=1, method="average")) - 1.0
        return np.multiply(np.abs(np.subtract(rd, self.instance, rd), rd),
                           self.__inst3, rd).mean()

    def __str__(self):
        """
        Get the name of this objective.

        :return: `"rankDifference{power}"`, where `{power}` is the power used
            for scaling the rank distance matrix.
        """
        return f"rankDifference{num_to_str(self.power)}"

    def lower_bound(self) -> float:
        """
        Get the lower bound: always `0`.

        :return: `0.0`
        """
        return 0.0

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("power", self.power)
        with logger.scope(SCOPE_INSTANCE) as scope:
            self.instance.log_parameters_to(scope)



"""
An improved bottom left encoding by Liu and Teng extended to multiple bins.

Here we provide an implementation of the improved bottom left encoding by Liu
and Teng [1], but extended to bins with limited height. If the height of the
bin is a limiting factor, then our implementation will automatically use
multiple bins. Another implementation is given in
:mod:`moptipyapps.binpacking2d.encodings.ibl_encoding_2`.

An :mod:`~moptipyapps.binpacking2d.instance` of the
two-dimensional bin packing problem defines a set of objects to be packed
and a bin size (width and height). Each object to be packed has itself a
width and a height as well as a repetition counter, which is `1` if the object
only occurs a single time and larger otherwise (i.e., if the repetition
counter is `5`, the object needs to be packaged five times).

The encoding receives signed permutations with repetitions as input. Each
element of the permutation identifies one object from the bin packing
instance. Each such object ID must occur exactly as often as the repetition
counter of the object in the instance data suggest. But the ID may also occur
negated, in which case the object is supposed to rotated by 90°.

Now our encoding processes such a permutation from beginning to end. It starts
with an empty bin `1`. Each object is first placed with its right end at the
right end of the bin and with its bottom line exactly at the top of the bin,
i.e., outside of the bin. Then, in each step, we move the object as far down
as possible. Then, we move it to the left as far as possible, but we
immediately stop if there was another chance to move the object down. In
other words, downward movements are preferred over left movements. This is
repeated until no movement of the object is possible anymore.

Once the object cannot be moved anymore, we check if it is fully inside the
bin. If yes, then the object is included in the bin and we continue with the
next object. If not, it does not fit into the bin.

This is the "Improved Bottom Left" heuristic by Liu and Teng [1].

If the object does not fit into the current bin, we place it at the
bottom-left corner of a new bin. We therefore increase the bin counter.
From now on, all the following objects will be placed into this bin until
the bin is full as well, in which case we move to the next bin again.
This means that the current bin is closed at the same moment the first
object is encountered that does not fit into it anymore. Therefore,
the objects in a closed bin do no longer need to be considered when packing
subsequent objects.

This is different from the second variant of this encoding implemented in file
:mod:`moptipyapps.binpacking2d.encodings.ibl_encoding_2`, which always checks
all the bins, starting at bin `1`, when placing any object. That other
encoding variant therefore must always consider all bins and is thus slower,
but tends to yield better packings.

This procedure has originally been developed and implemented by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所, http://iao.hfuu.edu.cn) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥学院) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).

1. Dequan Liu and Hongfei Teng. An Improved BL-Algorithm for Genetic Algorithm
   of the Orthogonal Packing of Rectangles. European Journal of Operational
   Research. 112(2):413-420. January (1999).
   https://doi.org/10.1016/S0377-2217(97)00437-2.
   http://www.paper.edu.cn/scholar/showpdf/MUT2AN0IOTD0Mxxh.
"""
from typing import Final

import numba  # type: ignore
import numpy as np
from moptipy.api.encoding import Encoding
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.types import type_error

from moptipyapps.binpacking2d.instance import (
    IDX_HEIGHT,
    IDX_WIDTH,
    Instance,
)
from moptipyapps.binpacking2d.packing import (
    IDX_BIN,
    IDX_BOTTOM_Y,
    IDX_ID,
    IDX_LEFT_X,
    IDX_RIGHT_X,
    IDX_TOP_Y,
    Packing,
)
from moptipyapps.shared import SCOPE_INSTANCE


@numba.njit(nogil=True, cache=True, inline="always", boundscheck=False)
def __move_down(packing: np.ndarray, bin_start: int, i1: int) -> bool:
    """
    Move the box at index `i1` down as far as possible in the current bin.

    `bin_start` is the index of the first object that has already been placed
    in the current bin. It always holds that `i1 >= bin_start`. In the case
    that `i1 == bin_start`, then we can move the object directly to the bottom
    of the bin without any issue.

    If `i1 > bin_start` we iterate over all objects at indices
    `bin_start...i1-1`. We first set `min_down` to the bottom-y coordinate of
    the box, because this is how far down we can move at most. Then, for each
    of the objects already placed in the bin, we check if there is any
    intersection of the horizontal with the current box. If there is no
    intersection *or* if the object is already above the current box, then the
    object will not influence the downward movement of our object. If there is
    an intersection, then we cannot move the current box deeper than the top-y
    coordinate of the other box.

    *Only* the box at index `i1` is modified and if it is modified, this
    function will return `True`.

    :param packing: the packing under construction
    :param bin_start: the starting index of the current bin
    :param i1: the index of the current box
    :return: `True` if the object was moved down, `False` if the object cannot
        be moved down any further because either it has reached the bottom or
        because it would intersect with other objects

    >>> # itemID, binID, left-x, bottom-y, right-x, top-y
    >>> r = np.array([[1, 1, 10, 20, 30, 40],
    ...               [2, 1, 30, 30, 50, 60],
    ...               [3, 1, 40, 100, 60, 200]])
    >>> __move_down(r, 0, 2)  # try to move down the box at index 2
    True
    >>> print(r[2, :])
    [  3   1  40  60  60 160]
    >>> __move_down(r, 0, 2)  # try to move down the box at index 2 again
    False
    >>> __move_down(r, 0, 1)  # try to move down the box at index 1 (ignore 2)
    True
    >>> print(r[1, :])
    [ 2  1 30  0 50 30]
    >>> __move_down(r, 0, 1)  # try to move down the box at index 1 again
    False
    >>> __move_down(r, 0, 2)  # try to move down the box at index 2 again now
    True
    >>> print(r[2, :])
    [  3   1  40  30  60 130]
    >>> __move_down(r, 0, 2)  # try to move down the box at index 2 again
    False
    >>> __move_down(r, 0, 0)
    True
    >>> print(r[0, :])
    [ 1  1 10  0 30 20]
    >>> __move_down(r, 0, 0)
    False
    """
    # load the coordinates of i1 into local variables to speed up computation
    packing_i1_left_x: Final[int] = int(packing[i1, IDX_LEFT_X])
    packing_i1_bottom_y: Final[int] = int(packing[i1, IDX_BOTTOM_Y])
    packing_i1_right_x: Final[int] = int(packing[i1, IDX_RIGHT_X])
    packing_i1_top_y: Final[int] = int(packing[i1, IDX_TOP_Y])
    min_down: int = packing_i1_bottom_y  # maximum move: down to bottom
    for i0 in range(bin_start, i1):  # iterate over all boxes in current bin
        # An intersection exists if the right-x of an existing box is larger
        # than the left-x of the new box AND if the left-x of the existing box
        # is less than the right-x of the new box.
        # Only intersections matter and only with objects not above us.
        if (packing[i0, IDX_RIGHT_X] > packing_i1_left_x) and \
                (packing[i0, IDX_LEFT_X] < packing_i1_right_x) and \
                (packing[i0, IDX_BOTTOM_Y] < packing_i1_top_y):
            # The object would horizontally intersect with the current object
            diff: int = int(packing_i1_bottom_y - packing[i0, IDX_TOP_Y])
            if diff < min_down:  # we can move down only a shorter distance
                min_down = diff  # ok, a shorter distance was found
    if min_down > 0:  # Can we move down? If yes, update box.
        packing[i1, IDX_BOTTOM_Y] = packing_i1_bottom_y - min_down
        packing[i1, IDX_TOP_Y] = packing_i1_top_y - min_down
        return True
    return False


@numba.njit(nogil=True, cache=True, inline="always", boundscheck=False)
def __move_left(packing: np.ndarray, bin_start: int, i1: int) -> bool:
    """
    Move the box at index `i1` left as far as possible in the current bin.

    This function moves a box to the left without changing its vertical
    position. It is slightly more tricky than the downwards moving function,
    because in the improved bottom left heuristic, downward moves are
    preferred compared to left moves. This means that the box needs to be
    stopped when reaching the edge of a box on whose top it sits.

    This function is to be called *after* `__move_down` and in an alternating
    fashion.

    *Only* the box at index `i1` is modified and if it is modified, this
    function will return `True`.

    :param packing: the packing under construction
    :param bin_start: the starting index of the current bin
    :param i1: the index of the current box
    :return: `True` if the object was moved down, `False` if the object cannot
        be moved down any further because either it has reached the bottom or
        because it would intersect with other objects

    >>> # itemID, binID, left-x, bottom-y, right-x, top-y
    >>> r = np.array([[1, 1,  0,  0, 30, 10],
    ...               [2, 1, 35,  0, 45, 30],
    ...               [3, 1,  0, 10, 10, 20],
    ...               [4, 1, 40, 30, 50, 40]])
    >>> __move_left(r, 0, 3)
    True
    >>> print(r[3, :])
    [ 4  1 25 30 35 40]
    >>> __move_left(r, 0, 3)
    True
    >>> print(r[3, :])
    [ 4  1  0 30 10 40]
    >>> r[3, :] = [4, 1, 25, 10, 35, 20]
    >>> __move_left(r, 0, 3)
    True
    >>> print(r[3, :])
    [ 4  1 10 10 20 20]
    >>> __move_left(r, 0, 3)
    False
    >>> # itemID, binID, left-x, bottom-y, right-x, top-y
    >>> r = np.array([[1, 1,  0,  0, 10, 2],
    ...               [2, 1, 10,  0, 20, 5],
    ...               [3, 1,  8,  2, 10, 4]])
    >>> __move_left(r, 0, 2)
    True
    >>> print(r[2, :])
    [3 1 0 2 2 4]
    """
    packing_i1_left_x: Final[int] = int(packing[i1, IDX_LEFT_X])
    packing_i1_bottom_y: Final[int] = int(packing[i1, IDX_BOTTOM_Y])
    packing_i1_right_x: Final[int] = int(packing[i1, IDX_RIGHT_X])
    packing_i1_top_y: Final[int] = int(packing[i1, IDX_TOP_Y])
    min_left: int = packing_i1_left_x
    for i0 in range(bin_start, i1):
        if packing[i0, IDX_LEFT_X] >= packing_i1_right_x:
            continue  # the object is already behind us, so it can be ignored
        if (packing[i0, IDX_RIGHT_X] > packing_i1_left_x) \
                and (packing[i0, IDX_LEFT_X] < packing_i1_right_x):
            # we have a horizontal intersection with a box below
            if packing[i0, IDX_TOP_Y] == packing_i1_bottom_y:
                # only consider those the box *directly* below and move the
                # right end of the new box to the left end of that box below
                diff = int(packing_i1_right_x - packing[i0, IDX_LEFT_X])
                if diff < min_left:
                    min_left = diff
        elif (packing_i1_top_y > packing[i0, IDX_BOTTOM_Y]) \
                and (packing_i1_bottom_y < packing[i0, IDX_TOP_Y]):
            diff = int(packing_i1_left_x - packing[i0, IDX_RIGHT_X])
            if diff < min_left:
                min_left = diff
    if min_left > 0:
        # move the box to the left
        packing[i1, IDX_LEFT_X] = packing_i1_left_x - min_left
        packing[i1, IDX_RIGHT_X] = packing_i1_right_x - min_left
        return True
    return False


@numba.njit(nogil=True, cache=True, inline="always", boundscheck=False)
def _decode(x: np.ndarray, y: np.ndarray, instance: np.ndarray,
            bin_width: int, bin_height: int) -> int:
    """
    Decode a (signed) permutation to a packing.

    The permutation is processed from the beginning to the end.
    Each element identifies one object by its ID. If the ID is negative,
    the object will be inserted rotated by 90°. If the ID is positive, the
    object will be inserted as is.

    The absolute value of the ID-1 will be used to look up the width and
    height of the object in the `instance` data. If the object needs to be
    rotated, width and height will be swapped.

    Each object is, at the beginning, placed with its right side at the right
    end of the bin. The bottom line of the object is initially put on top of
    the bin, i.e., initially the object is outside of the bin.

    Then, the object is iteratively moved downward as far as possible. Once it
    reaches another object, we move it to the left until either its right side
    reaches the left end of the object beneath it or until its left side
    touches another object. Then we try to move the object down again, and so
    on.

    Once the object can no longer be moved down, we check if it is now fully
    inside of the bin. If yes, then good, the object's bin index is set to the
    ID of the current bin. If not, then we cannot place the object into this
    bin. In this case, we increase the bin ID by one. The object is put into
    a new and empty bin. We move it to the bottom-left corner of this bin. In
    other words, the left side of the object touches the left side of the bin,
    i.e., is `0`. The bottom-line of the object is also the bottom of the bin,
    i.e., has coordinate `0` as well.

    All objects that are placed from now on will go into this bin until the
    bin is full. Then we move on to the next bin, and so on. In other words,
    once a bin is full, we no longer consider it for receiving any further
    objects.

    :param x: a possibly signed permutation
    :param y: the packing object
    :param instance: the packing instance data
    :param bin_width: the bin width
    :param bin_height: the bin height
    :returns: the number of bins

    As example, we use a slightly modified version (we add more objects so we
    get to see the use of a second bin) of Figure 2 of the Liu and Teng paper
    "An Improved BL-Algorithm for Genetic Algorithm of the Orthogonal Packing
    of Rectangles."

    >>> # [width, height, repetitions]
    >>> inst = np.array([[10, 20, 5], [5, 5, 5]])
    >>> # [id = plain, -id = rotated]
    >>> xx = np.array([1, -1, 2, -2, 1, -2, -2, -1, -1, 2])
    >>> # [id, bin, left-x, bottom-y, right-x, top-y] ...
    >>> yy = np.empty((10, 6), int)
    >>> print(_decode(xx, yy, inst, 30, 30))
    2
    >>> print(yy[0, :])
    [ 1  1  0  0 10 20]
    >>> print(yy[1, :])
    [ 1  1 10  0 30 10]
    >>> print(yy[2, :])
    [ 2  1 10 10 15 15]
    >>> print(yy[3, :])
    [ 2  1 15 10 20 15]
    >>> print(yy[4, :])
    [ 1  1 20 10 30 30]
    >>> print(yy[5, :])
    [ 2  1 10 15 15 20]
    >>> print(yy[6, :])
    [ 2  1 15 15 20 20]
    >>> print(yy[7, :])
    [ 1  1  0 20 20 30]
    >>> print(yy[8, :])
    [ 1  2  0  0 20 10]
    >>> print(yy[9, :])
    [ 2  2 20  0 25  5]
    """
    w: int  # the width of the current object
    h: int  # the height of the current object
    use_id: int  # the id of the current object
    bin_start: int = 0  # the index of the first object in the current bin
    bin_id: int = 1  # the id of the current bin
    for i, item_id in enumerate(x):  # iterate over all objects
        if item_id < 0:  # object should be rotated
            use_id = -(item_id + 1)  # get absolute id - 1
            w = int(instance[use_id, IDX_HEIGHT])  # width = height (rotated!)
            h = int(instance[use_id, IDX_WIDTH])   # height = width (rotated!)
        else:  # the object will not be rotated
            use_id = item_id - 1   # id - 1
            w = int(instance[use_id, IDX_WIDTH])  # get width
            h = int(instance[use_id, IDX_HEIGHT])  # get height

# It could be that an object is too wide or too high for the bin in its
# current rotation even if the bin was empty entirely. In this case, we simply
# force-rotate it. A bin packing instance will not permit objects that do not
# fit into the bin in any rotation. So if the object does not fit in its
# current rotation, it must fit if we simply rotate it by 90°.
        if (w > bin_width) or (h > bin_height):
            w, h = h, w

# At first, the object's right corner is at the right corner of the bin.
# The object sits exactly at the top of the bin, i.e., its bottom line
# is the top line of the bin.
        y[i, IDX_ID] = use_id + 1  # the id of the object
        y[i, IDX_LEFT_X] = bin_width - w  # the left end
        y[i, IDX_BOTTOM_Y] = bin_height  # object sits on top of bin
        y[i, IDX_RIGHT_X] = bin_width  # object ends at right end of bin
        y[i, IDX_TOP_Y] = bin_height + h  # top of object is outside of bin

        while __move_down(y, bin_start, i) or __move_left(y, bin_start, i):
            pass  # loop until object can no longer be moved

# If the object is not fully inside the current bin, we move to a new bin.
        if (y[i, IDX_RIGHT_X] > bin_width) or (y[i, IDX_TOP_Y] > bin_height):
            bin_id = bin_id + 1  # step to the next bin
            bin_start = i  # set the starting index of the bin
            y[i, IDX_LEFT_X] = 0  # the object goes to the left end of the bin
            y[i, IDX_BOTTOM_Y] = 0  # the object goes to the bottom of the bin
            y[i, IDX_RIGHT_X] = w  # so its right end is its width
            y[i, IDX_TOP_Y] = h  # and its top end is its height
        y[i, IDX_BIN] = bin_id  # store the bin id
    return int(bin_id)  # return the total number of bins


class ImprovedBottomLeftEncoding1(Encoding):
    """An Improved Bottem Left Encoding by Liu and Teng for multiple bins."""

    def __init__(self, instance: Instance) -> None:
        """
        Instantiate the improved best first encoding.

        :param instance: the packing instance
        """
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the internal instance reference
        self.__instance: Final[Instance] = instance

    def decode(self, x: np.ndarray, y: Packing) -> None:
        """
        Map a potentially signed permutation to a packing.

        :param x: the array
        :param y: the Gantt chart
        """
        y.n_bins = _decode(x, y, self.__instance, self.__instance.bin_width,
                           self.__instance.bin_height)

    def __str__(self) -> str:
        """
        Get the name of this encoding.

        :return: `"ibf1"`
        :rtype: str
        """
        return "ibf1"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        with logger.scope(SCOPE_INSTANCE) as kv:
            self.__instance.log_parameters_to(kv)



"""An instance of the ordering problem."""


from typing import Callable, Final, Iterable, TypeVar, cast

import numpy as np
from moptipy.api.component import Component
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import int_range_to_dtype
from moptipy.utils.types import check_int_range, check_to_int_range, type_error
from scipy.stats import rankdata  # type: ignore

#: the type variable for the source object
S = TypeVar("S")
#: the type variable for the object
OBJ = TypeVar("OBJ")


class Instance(Component, np.ndarray):
    """
    An instance of the One-Dimensional Ordering Problem.

    Such an instance represents the ranking of objects by their distance to
    each other. If we have `n` objects, then we compute the distance of each
    object to each of its `n-1` neighbors. The distance of an object to itself
    is `0`. Let's say that we have the three numbers `1, 2, 3` and the
    distance be their absolute difference. Then `1` has distance `1` to `2`
    and distance `2` to `3`. For `2`, the distance to `1` is `1` and to `3` is
    also `1`. We compute the average rank of the objects to each other on a
    per-row basis (including the object itself). For object `1`, this gives us
    `(1, 2, 3)` and for object `2`, it gives `(2.5, 1, 2.5)`, the latter
    because `1` and `3` are equally far from `2`. We multiply this with 2 to
    get integer ranks and subtract 1, i.e., we would get
    `[[1, 3, 5], [4, 1, 4], [5, 3, 1]]` for our three numbers.

    >>> def _dist(a, b):
    ...     return abs(a - b)
    >>> def _tags(a):
    ...     return f"t{a}"
    >>> the_instance = Instance.from_sequence_and_distance(
    ...     [1, 2, 3], _tags, _dist)
    >>> print(np.array(the_instance))
    [[1 3 5]
     [4 1 4]
     [5 3 1]]

    However, it is possible that some objects appear twice or have zero
    distance. In this case, they will be purged from the matrix:

    >>> the_instance = Instance.from_sequence_and_distance(
    ...     [1, 2, 3, 3, 2, 3], _tags, _dist)
    >>> print(np.array(the_instance))
    [[1 3 5]
     [4 1 4]
     [5 3 1]]

    But they will still be visible in the tags:

    >>> print(the_instance.tags)
    ((('t1',), 0), (('t2',), 1), (('t2',), 1), (('t3',), 2), \
(('t3',), 2), (('t3',), 2))
    """

    #: the assignment of string tags to object IDs
    tags: tuple[tuple[tuple[str, ...], int], ...]

    def __new__(cls, matrix: np.ndarray,
                tags: Iterable[tuple[Iterable[str] | str, int]]) -> "Instance":
        """
        Create an instance of the one-dimensional ordering problem.

        :param cls: the class
        :param matrix: the matrix with the rank data (will be copied)
        :param tags: the assignment of rows to names
        """
        n: Final[int] = len(matrix)
        use_shape: Final[tuple[int, int]] = (n, n)
        if isinstance(matrix, np.ndarray):
            if matrix.shape != use_shape:
                raise ValueError(
                    f"Unexpected shape {matrix.shape} for {n} objects, "
                    f"expected {use_shape}.")
        else:
            raise type_error(matrix, "matrix", np.ndarray)

        obj: Final[Instance] = super().__new__(
            cls, use_shape, int_range_to_dtype(min_value=-n, max_value=n))
        np.copyto(obj, matrix, "unsafe")
        for i in range(n):
            for j in range(n):
                if check_to_int_range(
                        obj[i, j], "element", 0, (2 * n) - 1) != matrix[i, j]:
                    raise ValueError(
                        f"error when copying: {obj[i, j]} != {matrix[i, j]}")

        #: the tags
        obj.tags = tuple(((t, ) if isinstance(t, str) else tuple(t),
                          k) for t, k in tags)
        if len(obj.tags) < n:
            raise ValueError(f"there must be at least {n} tags, but got "
                             f"{len(obj.tags)}, i.e., {obj.tags}")

        req_len: int = len(obj.tags[0][0])
        if req_len <= 0:
            raise ValueError(f"tag length must be >= 1, but is {req_len} "
                             f"for tag {obj.tags[0]}.")
        for tag in obj.tags:
            check_int_range(tag[1], "id", 0, n)
            if len(tag[0]) != req_len:
                raise ValueError(
                    f"all tags must have the same length. "
                    f"while the first tag ({obj.tags[0]}) has "
                    f"length {req_len}, {tag} has length {len(tag[0])}.")
        return obj

    @staticmethod
    def from_sequence_and_distance(
            data: Iterable[S | None],
            get_tags: Callable[[OBJ], str | Iterable[str]],
            get_distance: Callable[[OBJ, OBJ], int],
            unpack: Callable[[S], OBJ | None] = cast(
                Callable[[S], OBJ], lambda x: x)) -> "Instance":
        """
        Turn a sequence of objects into a One-Dimensional Ordering instance.

        :param data: the data source, i.e., an iterable of data elements
        :param get_tags: the function for extracting tags from objects
        :param get_distance: the function for getting the distance between
            objects
        :param unpack: a function unpacking a source element, e.g., a string,
            to an object
        :return: the ordering instance

        >>> def _dist(a, b):
        ...     return abs(a - b)
        >>> def _tags(a):
        ...     return f"x{a}", f"b{a}"
        >>> res = Instance.from_sequence_and_distance(
        ...     [4, 5, 1, 2, 3], _tags, _dist)
        >>> print(res)
        Instance
        >>> print(np.array(res))
        [[1 4 9 7 4]
         [3 1 9 7 5]
         [7 9 1 3 5]
         [7 9 4 1 4]
         [4 8 8 4 1]]
        >>> print(res.tags)
        ((('x4', 'b4'), 0), (('x5', 'b5'), 1), (('x1', 'b1'), 2), \
(('x2', 'b2'), 3), (('x3', 'b3'), 4))
        >>> res = Instance.from_sequence_and_distance(
        ...     [4, 5, 4, 2, 5], _tags, _dist)
        >>> print(np.array(res))
        [[1 3 5]
         [3 1 5]
         [3 5 1]]
        >>> print(res.tags)
        ((('x4', 'b4'), 0), (('x4', 'b4'), 0), (('x5', 'b5'), 1), \
(('x5', 'b5'), 1), (('x2', 'b2'), 2))
        >>> def _dist2(a, b):
        ...     return abs(abs(a) - abs(b))
        >>> res = Instance.from_sequence_and_distance(
        ...     [-4, 5, 4, 2, -5], _tags, _dist2)
        >>> print(np.array(res))
        [[1 3 5]
         [3 1 5]
         [3 5 1]]
        >>> print(res.tags)
        ((('x4', 'b4'), 0), (('x-4', 'b-4'), 0), (('x-5', 'b-5'), 1), \
(('x5', 'b5'), 1), (('x2', 'b2'), 2))
        """
        if not isinstance(data, Iterable):
            raise type_error(data, "data", Iterable)
        if not callable(get_tags):
            raise type_error(get_tags, "get_tags", call=True)
        if not callable(get_distance):
            raise type_error(get_distance, "get_distance", call=True)
        if not callable(unpack):
            raise type_error(unpack, "unpack", call=True)

        # first extract all the objects
        raw: list[OBJ] = []
        for d in data:
            if d is None:
                continue
            objx: OBJ | None = unpack(d)
            if objx is None:
                continue
            raw.append(objx)

        # build a distance matrix and purge all zero-distance elements
        mappings: list[tuple[OBJ, int]] = []
        distances: list[list[int]] = []
        i: int = 0
        while i < len(raw):
            o1: OBJ = raw[i]
            j: int = i + 1
            current_dists: list[int] = [d[i] for d in distances]
            current_dists.append(0)
            while j < len(raw):
                o2: OBJ = raw[j]
                dist: int = check_int_range(
                    get_distance(o1, o2), "get_distance", 0,
                    1_000_000_000_000)
                if dist <= 0:  # distance == 0, must purge
                    mappings.append((o2, i))
                    del raw[j]
                    for ds in distances:
                        del ds[j]
                    continue
                current_dists.append(dist)
                j += 1
            mappings.append((o1, i))
            distances.append(current_dists)
            i += 1

        # we now got a full distance matrix, let's turn it into a rank matrix
        return Instance(
            (2.0 * rankdata(distances, axis=1, method="average") - 1.0),
            ((get_tags(obj), idx) for obj, idx in mappings))

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("nCleaned", len(self))
        logger.key_value("nOrig", len(self.tags))



"""
An objective functions for the dynamic control problem.

The dynamic control problem means that our system starts in a given state
and we try to move it to a stable state by using control effort. Controllers
are trained over several training states, for each of which we can compute a
figure of merit.

We offer two different approaches for this:

- :class:`FigureOfMerit` computes the arithmetic mean `z` over the separate
  figures of merit `J` of the training cases.
- :class:`FigureOfMeritLE` tries to smooth out the impact of bad starting
  states by computing `exp(mean[log(J + 1)]) - 1`.

These objective functions also offer a way to collect the state+control and
corresponding differential vectors.
"""

from math import expm1
from typing import Callable, Final

import numpy as np
from moptipy.api.objective import Objective
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.nputils import array_to_str
from moptipy.utils.types import type_error

from moptipyapps.dynamic_control.instance import Instance
from moptipyapps.dynamic_control.ode import diff_from_ode, j_from_ode, run_ode
from moptipyapps.shared import SCOPE_INSTANCE


class FigureOfMerit(Objective):
    """A base class for figures of merit."""

    def __init__(self, instance: Instance,
                 supports_model_mode: bool = False) -> None:
        """
        Create the figure-of-merit objective of the dynamic control problem.

        :param instance: the instance
        :param supports_model_mode: `True` if this objective is supposed to
            support alternating actual and model-based runs, `False` if it is
            just applied to the actual instance (see :meth:`set_model` and
            :meth:`get_differentials`).
        """
        super().__init__()
        if not isinstance(instance, Instance):
            raise type_error(instance, "instance", Instance)
        #: the dynamic control instance
        self.instance: Final[Instance] = instance
        #: the simulation steps
        self.__steps: Final[int] = instance.system.training_steps
        #: the training time
        self.__time: Final[float] = instance.system.training_time
        #: the training starting states
        self.__training: Final[np.ndarray] = \
            instance.system.training_starting_states
        #: the results
        self.__results: Final[np.ndarray] = np.empty(
            len(self.__training), float)
        #: the equations
        self.__equations: Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None] = \
            instance.system.equations
        #: the controller
        self.__controller: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] = \
            instance.controller.controller
        #: the controller dimension
        self.__controller_dim: Final[int] = instance.controller.control_dims
        #: the collection of state+control and differential vectors
        self.__collection: Final[list[tuple[np.ndarray, np.ndarray]] | None] \
            = [] if supports_model_mode else None
        #: should we collect data?
        self.__collect: bool = self.__collection is not None
        #: the state dimensions inside the `J`
        self.__state_dims_in_j: Final[int] = instance.system.state_dims_in_j
        #: the weight of the control effort
        self.__gamma: Final[float] = instance.system.gamma

    def initialize(self) -> None:
        """Initialize the objective for use."""
        super().initialize()
        if self.__collection is not None:
            self.__collection.clear()
        self.set_raw()

    def set_raw(self) -> None:
        """
        Let this objective work on the original system equations.

        The objective function here can be used in two modi: a) based on the
        original systems model, as given in
        :attr:`~moptipyapps.dynamic_control.instance.Instance.system`, or b)
        on a learned model of the system. This function here toggles to the
        former mode, i.e., to the actual system mode. In this modus, training
        data for training the system model will be gathered if the objective
        function is configured to do so. In that case, you can toggle to model
        mode via :meth:`set_model`.
        """
        self.__equations = self.instance.system.equations
        self.__collect = self.__collection is not None

    def get_differentials(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Get the collected differentials.

        If `supports_model_mode` was set to `True` in the creating of this
        objective function, then the system will gather tuples `(s, c)` and
        `ds/dt` when in raw mode (see :meth:`set_raw`) and make them available
        here to train system models (see :meth:`set_model`). Notice that
        gathering training data is a very memory intense process.

        :returns: the collected differentials
        """
        cl: Final[list[tuple[np.ndarray, np.ndarray]] | None] = \
            self.__collection
        if cl is None:
            raise ValueError("Differential collection not supported.")
        if len(cl) == 1:
            return cl[0]  # pylint: disable=E1136
        result = (np.concatenate([t[0] for t in cl]),  # pylint: disable=E1133
                  np.concatenate([t[1] for t in cl]))  # pylint: disable=E1133
        cl.clear()
        cl.append(result)
        return result

    def set_model(self, equations: Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]) -> None:
        """
        Set the model-driven mode for the evaluation.

        In this modus, the internal system equations are replaced by the
        callable `equations` passed into this function and the data collection
        is stopped. The idea is that `equations` could be a model synthesized
        on the data gathered (see :meth:`get_differentials`) and thus does not
        represent the actual dynamic system but a model thereof. We could
        synthesize a controller for this model and for this purpose would use
        the exactly same objective function -- just instead of using the
        actual system equations, we use the system model. Of course, we then
        need to deactivate the data gathering mechanism (see again
        :meth:`get_differentials`), because the data would then not be real
        system data. You can toggle back to the actual system using
        :meth:`set_raw`.

        :param equations: the equations to be used instead of the actual
            system's differential equations.
        """
        if self.__collection is None:
            raise ValueError("Cannot go into model mode without gathering "
                             "model training data!")
        self.__equations = equations
        self.__collect = False

    def __str__(self) -> str:
        """
        Get the name of this objective function.

        :return: `figureOfMerit`
        """
        return "figureOfMerit"

    def evaluate(self, x: np.ndarray) -> float:
        """
        Evaluate the parameterization of a controller.

        :param x: the controller parameters
        :return: the figure of merit
        """
        steps: Final[int] = self.__steps
        time: Final[float] = self.__time
        training: Final[np.ndarray] = self.__training
        results: Final[np.ndarray] = self.__results
        equations: Final[Callable[[
            np.ndarray, float, np.ndarray, np.ndarray], None]] \
            = self.__equations
        controller: Final[
            Callable[[np.ndarray, float, np.ndarray, np.ndarray], None]] \
            = self.__controller
        controller_dim: Final[int] = self.__controller_dim
        collector: Final[Callable[[tuple[
            np.ndarray, np.ndarray]], None] | None] = \
            self.__collection.append if self.__collect else None
        state_dim: Final[int] = len(training[0])
        state_dims_in_j: Final[int] = self.__state_dims_in_j
        gamma: Final[float] = self.__gamma

        for i, start in enumerate(training):
            # The following line makes no sense at all. It creates a copy of
            # the flattened version of the (already flat) start. The copy is
            # stored nowhere, so it is immediately discarded. The value of
            # start is not changed. However, the numpy array container
            # changes, for an unclear reason. This is required and it must
            # happen exactly here, for an unclear reason. Otherwise, the
            # results of the objective function are inconsistent. For an
            # unclear reason.
            np.copy(start.flatten())  # <--- This should make no sense...
            ode = run_ode(
                start, equations, controller, x, controller_dim, steps, time)
            results[i] = z = j_from_ode(ode, state_dim, state_dims_in_j, gamma)
            if not (0.0 <= z <= 1e100):
                return 1e200
            if collector is not None:
                collector(diff_from_ode(ode, state_dim))
        z = self.sum_up_results(results)
        return z if 0.0 <= z <= 1e100 else 1e200

    def sum_up_results(self, results: np.ndarray) -> float:
        """
        Compute the final objective value from several single `J` values.

        When synthesizing controllers, we do not just apply them to a single
        simulation run. Instead, we use multiple training cases (see
        :attr:`~moptipyapps.dynamic_control.system.System.\
training_starting_states`) and perform :attr:`~moptipyapps.dynamic_control\
.system.System.training_steps` simulation steps on each of them. Each such
        training starting state will result in a single `J` value, which is
        the sum of squared state and control values. We now compute the end
        objective value from these different `J` values by using this
        function here.

        This will *destroy* the contents of `results`.

        :param results: the array of `J` values
        :return: the final result
        """
        return float(results.mean())

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log all parameters of this component as key-value pairs.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("modelModeEnabled", self.__collection is not None)
        logger.key_value("dataCollecting", self.__collect)
        eq: Final = self.__equations
        logger.key_value("usingOriginalEquations",
                         eq is self.instance.system.equations)
        mp: Final[str] = "modelParameters"
        if hasattr(self.__equations, mp):
            logger.key_value(mp, array_to_str(getattr(eq, mp)))
        with logger.scope(SCOPE_INSTANCE) as scope:
            self.instance.log_parameters_to(scope)

    def lower_bound(self) -> float:
        """
        Get the lower bound of the figure of merit, which is 0.

        :returns: 0.0
        """
        return 0.0


class FigureOfMeritLE(FigureOfMerit):
    """
    Compute a `exp(mean(log(z+1)))-1` over the figures of merit `z`.

    Different from :class:`FigureOfMerit`, we compute the mean of `log(z + 1)`
    where `z` be the figures of merit of the single training cases. We then
    return `exp(mean[log(z + 1)]) - 1` as final result. The goal is to reduce
    the impact of training cases that require more control effort.

    If we solve the dynamic control problem for diverse training cases, then
    we may have some very easy cases, where the system just needs a small
    control impulse to move into a stable and cheap state. Others may have
    very far out and expensive starting states that require lots of control
    efforts to be corrected. If we simply average over all states, then these
    expensive states will dominate whatever good we are doing in the cheap
    states. Averaging over the `log(J+1)` reduces such impact. We then compute
    `exp[...]-1` of the result as cosmetics to get back into the original
    range of the figure of merits.
    """

    def __str__(self) -> str:
        """
        Get the name of this objective function.

        :return: `figureOfMeritLE`
        """
        return "figureOfMeritLE"

    def sum_up_results(self, results: np.ndarray) -> float:
        """
        Compute the final objective value from several single `J` values.

        For each training case, there is one basic figure of merit `J` and
        here we compute `exp(mean[log(J + 1)]) - 1` over all of these values.

        :param results: the array of `J` values
        :return: the final result
        """
        return float(expm1(np.log1p(results, results).mean()))



"""
Applications of Metaheuristic Optimization in Python.

Currently, the following applications are implemented:

- :mod:`~moptipyapps.binpacking2d` provides methods to solve two-dimensional
  bin packing instances,
- :mod:`~moptipyapps.dynamic_control` can simulate controlled dynamic systems
  of differential equations and our goal is to synthesize controllers that can
  steer such systems into beneficial states,
- mod:`~moptipyapps.tsp` offers instances of the well-known Traveling
  Salesperson Problem (TSP) and some very basic algorithms to tackle it.

The following additional tools are implemented:

- :mod:`~moptipyapps.tests` offers unit tests to try out optimization
  algorithms and other instances of :mod:`~moptipy.api.component` on the
  different problems that are provided above.
- :mod:`~moptipyapps.shared` offers shared constants and tools.
"""

# During some local pip installs, the other already installed packages may not
# be accessible. This may cause
# `from moptipy.utils.sys_info import add_dependency` to fail during
# `pip3 --no-input --timeout 360 --retries 100 -v install .`.
# The reason is not clear to me.
# To prevent this issue, we now try to load the `moptipy` module and catch the
# corresponding `ModuleNotFoundError` exception. Only if the module is loaded
# correctly, we add `moptipyapps` to the dependencies.
# Of course, this ignores the situation where there would really be an error.
# But that is OK, because in that case we will crash later anyway.

can_do: bool = True
try:
    from moptipy.utils.sys_info import add_dependency
except ModuleNotFoundError:
    can_do = False

if can_do:
    add_dependency("moptipyapps", ignore_if_make_build=True)



"""
An illustrator for results gathered from ODE integration to multi-plots.

This evaluator will create figures where each sub-plot corresponds to one
evolution of the system over time. The starting state is marked with a blue
cross, the final state with a red one. Both states are connected with a line
that marks the progress over time. The line begins with blue color and changes
its color over violet/pinkish towards yellow the farther in time the system
progresses.
"""

from contextlib import AbstractContextManager
from math import isfinite, sqrt
from os.path import basename, dirname
from typing import Any, Callable, Collection, Final

import matplotlib as mpl  # type: ignore
import moptipy.utils.plot_utils as pu
import numpy as np
from matplotlib.axes import Axes  # type: ignore
from matplotlib.figure import Figure  # type: ignore
from matplotlib.lines import Line2D  # type: ignore
from moptipy.utils.console import logger
from moptipy.utils.path import Path
from moptipy.utils.strings import float_to_str
from moptipy.utils.types import check_int_range
from mpl_toolkits.mplot3d.art3d import Line3D  # type: ignore


def _str(f: float) -> str:
    """
    Convert a floating point number to a short string.

    :param f: the number
    :return: the string
    """
    f = float(f)
    a1 = float_to_str(f)
    a2 = f"{f:.2f}"
    while a2.endswith("0"):
        a2 = a2[:-1]
    if a2.endswith("."):
        a2 = a2[:-1]
    if (a1 == "-0") or (a2 == "-0"):
        return "0"
    return a1 if len(a1) < len(a2) else a2


def _vec(v: np.ndarray) -> str:
    """
    Convert a vector to a quick string.

    :param v: the vector
    :return: the string
    """
    s = ",".join(_str(vv) for vv in v)
    return f"({s})"


#: the color maps to use
__COLOR_MAPS: Final[tuple[str, str, str, str, str, str, str, str, str]] = (
    "spring", "winter", "copper", "plasma", "cool", "hsv", "winter",
    "coolwarm", "BrBG")


def _get_colors(alen: int, cmi: int = 0, multi: bool = False) -> Callable[
        [float], tuple[float, float, float, float]]:
    """
    Get a given color map with the given number of columns.

    For a given color map index (`cmi`), this function returns a callable that
    maps values from `[0,1]` to colors. If this is supposed to be the only
    color map in use (`m̀ulti == False`), then it will return the standard
    `spring` color map. Otherwise, it will return different color maps
    depending on the color map index `cmi`.

    :param alen: the number of colors
    :param cmi: the color map name index
    :param multi: are there multiple colors
    :return: the colors
    """
    if multi:
        if cmi == 0:  # pink
            return lambda x: (0.18 + 0.8 * x, 0.0, 0.18 + 0.8 * x, 1.0)
        if cmi == 1:  # green
            return lambda x: (0.0, 0.18 + 0.8 * x, 0.0, 1.0)
        if cmi == 2:  # blue
            return lambda x: (0.0, 0.0, 0.18 + 0.8 * x, 1.0)
        cmi -= 2

    cm: Final[str] = __COLOR_MAPS[
        check_int_range(cmi, "color_map_index", 0, 10000) % len(__COLOR_MAPS)]
    obj: Final = _get_colors

    if hasattr(obj, cm):
        the_colors: dict[int, Callable[[
            float], tuple[float, float, float, float]]] = getattr(obj, cm)
    else:
        the_colors = {}
        setattr(obj, cm, the_colors)
    if alen in the_colors:
        return the_colors[alen]
    colors: Callable[[float], tuple[float, float, float, float]] \
        = mpl.colormaps[cm].resampled(alen)
    the_colors[alen] = colors
    return colors


#: the start point color
START_COLOR: Final[str] = "blue"
#: the start marker
START_MARKER: Final[str] = "++"
#: the end point color
END_COLOR: Final[str] = "red"
#: the end marker
END_MARKER: Final[str] = "o\u25CF"


def _plot_2d(state: np.ndarray, axes: Axes,
             xi: int, yi: int,
             colors: Callable[[float], tuple[float, float, float, float]],
             z: int) -> int:
    """
    Plot a figure in 2D.

    :param state: the state matrix
    :param axes: the plot
    :param xi: the x-index
    :param yi: the y-index
    :param colors: the colors to use
    :param z: the z index
    :returns: the new z index
    """
    size: Final[int] = len(state) - 1
    for i in range(size):
        z += 1
        axes.add_artist(Line2D(state[i:i + 2, xi],
                               state[i:i + 2, yi],
                        color=colors(i / size), zorder=z))
    z += 1
    axes.scatter(state[0, xi], state[0, yi], color=START_COLOR,
                 zorder=z, marker=START_MARKER[0])
    z += 1
    axes.scatter(state[-1, xi], state[-1, yi], color=END_COLOR,
                 zorder=z, marker=END_MARKER[0])
    return z + 1


def _plot_3d(state: np.ndarray, axes: Axes,
             xi: int, yi: int, zi: int,
             colors: Callable[[float], tuple[float, float, float, float]],
             z: int, ranges: list[float]) -> int:
    """
    Plot a figure in 3D.

    :param state: the state matrix
    :param axes: the plot
    :param xi: the x-index
    :param yi: the y-index
    :param yi: the z-index
    :param colors: the colors to use
    :param z: the z index
    :param ranges: the axes ranges
    :returns: the new z index
    """
    if hasattr(axes, "force_zorder"):
        axes.force_zorder = True

    size: Final[int] = len(state) - 1
    for i in range(size):
        z += 1
        axes.add_artist(Line3D(state[i:i + 2, xi],
                               state[i:i + 2, yi],
                               state[i:i + 2, zi],
                        color=colors(i / size), zorder=z))

    col: str = START_COLOR
    for p in [state[0], state[-1]]:
        w: float = 0.04 * ranges[0]
        z += 1
        axes.add_artist(Line3D((p[xi] - w, p[xi] + w),
                               (p[yi], p[yi]),
                               (p[zi], p[zi]),
                        color=col, zorder=z))
        z += 1
        w = 0.04 * ranges[1]
        axes.add_artist(Line3D((p[xi], p[xi]),
                               (p[yi] - w, p[yi] + w),
                               (p[zi], p[zi]),
                        color=col, zorder=z))
        z += 1
        w = 0.04 * ranges[2]
        axes.add_artist(Line3D((p[xi], p[xi]),
                               (p[yi], p[yi]),
                               (p[zi] - w, p[zi] + w),
                        color=col, zorder=z))
        col = END_COLOR
    return z + 1


class ResultsPlot(AbstractContextManager):
    """
    A class for plotting results via `multi_run_ode`.

    Function :func:`moptipyapps.dynamic_control.ode.multi_run_ode` can pass
    its results to various output generating procedures. This class here
    offers a procedure for plotting them to a file.
    """

    def __init__(self, dest_file: str, sup_title: str | None,
                 state_dims: int,
                 plot_indices: Collection[int] = (0, ),
                 state_dim_mod: int = 0) -> None:
        """
        Create the ODE plotter.

        :param dest_file: the path to the destination file
        :param sup_title: the title for the figure
        :param state_dims: the state dimensions
        :param plot_indices: the indices that are supposed to be plotted
        :param state_dim_mod: the modulus for the state dimensions
        """
        super().__init__()
        #: the destination file
        self.__dest_file: Final[Path] = Path.path(dest_file)
        logger(f"plotting data to file {self.__dest_file!r}.")
        #: the state dimensions
        self.__state_dims: Final[int] = check_int_range(
            state_dims, "state_dimes", 2, 1_000_000)
        #: The modulus for the state dimensions for plotting
        self.__state_dim_mod: Final[int] = check_int_range(
            state_dim_mod, "state_dim_mod", 0, state_dims)

        total_plots: Final[int] = len(plot_indices)
        #: the plot indexes
        if not (0 < total_plots <= 9):
            raise ValueError(f"Invalid plot indices {plot_indices!r}, "
                             f"contains {total_plots} elements.")
        #: the plot indices
        self.__plot_indices: Final[Collection[int]] = plot_indices

        srt: Final[int] = max(3, int(round(1 + sqrt(total_plots))))
        args: Final[dict[str, Any]] = {
            "items": total_plots,
            "max_items_per_plot": 1,
            "max_rows": srt,
            "max_cols": srt,
        }
        use_dims: Final[int] = state_dims if state_dim_mod <= 0 \
            else state_dim_mod
        if use_dims >= 3:
            args["plot_config"] = {"projection": "3d"}

        figure, plots = pu.create_figure_with_subplots(**args)
        #: the figure
        self.__figure: Figure | None = figure
        #: the plots
        self.__subplots: None | tuple[tuple[
            Axes | Figure, int, int, int, int, int], ...] = plots
        #: the current plot
        self.__cur_plot: int = 0
        #: the z-order
        self.__z: int = 0

        if sup_title is not None:
            figure.suptitle(sup_title)

    def collector(self, index: int, ode: np.ndarray,
                  j: float, time: float) -> None:
        """
        Plot the result of a multi-ode run.

        :param index: the index of the result
        :param ode: the ode result matrix
        :param j: the figure of merit
        :param time: the time value
        """
        if self.__figure is None:
            raise ValueError("Already closed output figure!")
        if index not in self.__plot_indices:
            return

        axes: Final[Axes] = pu.get_axes(
            self.__subplots[self.__cur_plot][0])
        state_dims: Final[int] = self.__state_dims
        state_dim_mod: Final[int] = self.__state_dim_mod
        use_dims: Final[int] = state_dims if state_dim_mod <= 0 \
            else state_dim_mod

        # make the appropriate title
        v1: str = _vec(ode[0, 0:state_dims])
        v2: str = _vec(ode[-1, 0:state_dims])
        if use_dims == 2:
            v1 = f"{START_MARKER[1]}{v1}"
            v2 = f"{END_MARKER[1]}{v2}"
        lv1: Final[int] = len(v1)
        lv2: Final[int] = len(v2)
        title: str = f"{v1}\u2192{v2}" if ((lv1 + lv2) < 40) else (
            f"{v1}\u2192\n{v2}" if lv1 <= lv2 else f"{v1}\n\u2192{v2}")
        title = f"{title}\nJ={j}, T={time:.4f}"

        state_min: Final[list[float]] = [
            ode[:, i:state_dims:state_dim_mod].min() for i in range(use_dims)]
        state_max: Final[list[float]] = [
            ode[:, i:state_dims:state_dim_mod].max() for i in range(use_dims)]
        ranges: Final[list[float]] = [
            state_max[i] - state_min[i] for i in range(use_dims)]

        can_plot: bool = all(
            isfinite(state_min[i]) and (state_min[i] > -1e20)
            and isfinite(state_max[i]) and (state_max[i] < 1e20)
            and (state_min[i] < state_max[i]) and isfinite(ranges[i])
            and (0 < ranges[i] < 1e20) for i in range(use_dims))

        if can_plot:
            if (max(1e-20, min(ranges)) / max(ranges)) > 0.25:
                axes.set_aspect("equal")
            axes.set_xlim(state_min[0], state_max[0])
            axes.set_ylim(state_min[1], state_max[1])
            if (use_dims > 2) and hasattr(axes, "set_zlim"):
                axes.set_zlim(state_min[2], state_max[2])

            size: Final[int] = len(ode) - 2
            for ci, i in enumerate(range(0, state_dims, use_dims)):
                colors: Callable[[float], tuple[
                    float, float, float, float]] = _get_colors(
                    size, ci, 0 < state_dim_mod < state_dims)
                if use_dims == 3:
                    self.__z = _plot_3d(ode, axes, i, i + 1, i + 2, colors,
                                        self.__z, ranges)
                elif use_dims == 2:
                    self.__z = _plot_2d(ode, axes, i, i + 1, colors,
                                        self.__z)
                else:
                    raise ValueError(f"Huh? state_dim={state_dims}, "
                                     f"state_dim_mod={state_dim_mod}??")
        else:
            title = f"{title}\n<cannot plot>"

        axes.set_title(title)
        self.__cur_plot += 1

    def __exit__(self, _, __, ___) -> None:
        """
        Close this context manager.

        :param _: the exception type; ignored
        :param __: the exception value; ignored
        :param ___: the exception whatever; ignored
        """
        if self.__figure is not None:
            try:
                the_dir: Final[str] = dirname(self.__dest_file)
                the_file: Final[str] = basename(self.__dest_file)
                dot: Final[int] = the_file.rindex(".")
                prefix: Final[str] = the_file[:dot]
                suffix: Final[str] = the_file[dot + 1:]
                pu.save_figure(self.__figure, prefix, the_dir, suffix)
                logger(f"finished plotting data to {self.__dest_file!r}.")
            finally:
                self.__figure = None
                self.__subplots = None



"""
A set of tools to load known optima in the TSPLib format.

The TSPLib provides benchmark instances for the Traveling Salesperson
Problem (TSP). We provide a subset of these instances with up to 2000 cities
as resources in :mod:`~moptipyapps.tsp.instance`. Additionally, TSPLib
provides the optimal solutions for a subset of these instances. Within this
module here, we provide methods for reading the optimal solution files
in the TSPLib format. We also include as resources the optimal solutions to
the instances that our package provide as resources as well.

You can get the list of instance names for which the optimal tours are
included in this package via :func:`list_resource_tours` and then load a tour
from a resource via :func:`opt_tour_from_resource`. If you want to read a tour
from an external file that obeys the TSPLib format, you can do so via
:func:`opt_tour_from_file`.

1. Gerhard Reinelt. TSPLIB - A Traveling Salesman Problem Library.
   *ORSA Journal on Computing* 3(4):376-384. November 1991.
   https://doi.org/10.1287/ijoc.3.4.376.
   http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/
2. Gerhard Reinelt. *TSPLIB95.* Heidelberg, Germany: Universität
   Heidelberg, Institut für Angewandte Mathematik. 1995.
   http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/tsp95.pdf
"""

from typing import Final, TextIO, cast

import moptipy.utils.nputils as npu
import numpy as np
from moptipy.utils.path import Path
from moptipy.utils.types import check_to_int_range, type_error

from moptipyapps.tsp.tsplib import open_resource_stream

#: the set of known optimal tours
_TOURS: Final[tuple[str, ...]] = (
    "a280", "att48", "bayg29", "bays29", "berlin52", "brg180", "ch130",
    "ch150", "eil101", "eil51", "eil76", "fri26", "gr120", "gr202", "gr24",
    "gr48", "gr666", "gr96", "kroA100", "kroC100", "kroD100", "lin105",
    "pcb442", "pr1002", "pr76", "rd100", "st70", "tsp225", "ulysses16",
    "ulysses22")


def _from_stream(stream: TextIO) -> np.ndarray:
    """
    Read an optimal tour from a stream.

    :param stream: the text stream
    :return: the tour
    """
    nodes: list[int] = []
    in_tour: bool = False
    done_nodes: set[int] = set()
    max_node: int = -1

    for the_line in stream:
        if not isinstance(the_line, str):
            raise type_error(the_line, "line", str)
        line = the_line.strip().upper()
        if len(line) <= 0:
            continue
        if line == "TOUR_SECTION":
            in_tour = True
            continue
        if line in ("-1", "EOF"):
            break
        if in_tour:
            for node_str in line.rsplit():
                node = check_to_int_range(
                    node_str, "node", 1, 1_000_000_000_000)
                if node in done_nodes:
                    raise ValueError(f"encountered node {node} twice")
                done_nodes.add(node)
                if node > max_node:
                    max_node = node
                nodes.append(node - 1)
    if len(nodes) != max_node:
        raise ValueError(
            f"expected {max_node} nodes, got {len(nodes)} instead.")
    return np.array(nodes, npu.int_range_to_dtype(0, max_node - 1))


def opt_tour_from_file(path: str) -> np.ndarray:
    """
    Read a known optimal tour from a file.

    :param path: the path to the file
    :return: the tour
    """
    file: Final[Path] = Path.file(path)
    with file.open_for_read() as stream:
        try:
            return _from_stream(cast(TextIO, stream))
        except (TypeError, ValueError) as err:
            raise ValueError(f"error when parsing file {file!r}") from err


def opt_tour_from_resource(name: str) -> np.ndarray:
    """
    Load an optimal tour from a resource.

    >>> np.array2string(opt_tour_from_resource("ulysses16"))
    '[ 0 13 12 11  6  5 14  4 10  8  9 15  2  1  3  7]'

    :param name: the name string
    :return: the optimal tour
    """
    if not isinstance(name, str):
        raise type_error(name, "name", str)
    with open_resource_stream(f"{name}.opt.tour") as stream:
        return _from_stream(stream)


def list_resource_tours() -> tuple[str, ...]:
    """
    Get a tuple of the names of the optimal tours in the resources.

    >>> list_resource_tours()
    ('a280', 'att48', 'bayg29', 'bays29', 'berlin52', 'brg180', 'ch130', \
'ch150', 'eil101', 'eil51', 'eil76', 'fri26', 'gr120', 'gr202', 'gr24', \
'gr48', 'gr666', 'gr96', 'kroA100', 'kroC100', 'kroD100', 'lin105', \
'pcb442', 'pr1002', 'pr76', 'rd100', 'st70', 'tsp225', 'ulysses16', \
'ulysses22')

    :return: the tuple
    """
    return _TOURS



"""Perform tests on the Traveling Salesperson Problem."""

from typing import Callable, Final, Iterable

import numpy as np
from moptipy.api.algorithm import Algorithm
from moptipy.api.objective import Objective
from moptipy.spaces.permutations import Permutations
from moptipy.tests.algorithm import validate_algorithm
from moptipy.tests.objective import validate_objective
from moptipy.utils.types import type_error
from numpy.random import Generator, default_rng

from moptipyapps.tsp.instance import Instance
from moptipyapps.tsp.tour_length import TourLength

#: the internal random number generator
__RANDOM: Final[Generator] = default_rng()


def tsp_instances_for_tests(
        random: Generator = __RANDOM,
        symmetric: bool = True,
        asymmetric: bool = True) -> Iterable[str]:
    """
    Get a sequence of TSP instances to test on.

    :param random: the random number generator to use
    :param symmetric: include symmetric instances
    :param asymmetric: include asymmetric instances
    :returns: an iterable of TSP instance names
    """
    if not isinstance(symmetric, bool):
        raise type_error(symmetric, "symmetric", bool)
    if not isinstance(asymmetric, bool):
        raise type_error(asymmetric, "asymmetric", bool)

    instances: tuple[str, ...]
    if symmetric and asymmetric:
        instances = Instance.list_resources(True, True)
    elif symmetric:
        instances = Instance.list_resources(True, False)
    elif asymmetric:
        instances = Instance.list_resources(False, True)
    else:
        raise ValueError(
            "at least of one symmetric or asymmetric must be TRUE")
    use_insts = list(instances)
    while len(use_insts) > 20:
        del use_insts[random.integers(len(use_insts))]
    random.shuffle(use_insts)
    return use_insts


def make_tour_valid(random: Generator, y: np.ndarray) -> np.ndarray:
    """
    Create valid TSP tours.

    :param random: the random number generator to use
    :param y: the input tour
    :returns: the valid version of `y`
    """
    y[0:len(y)] = range(len(y))
    random.shuffle(y)
    return y


def make_tour_invalid(random: Generator, y: np.ndarray) -> np.ndarray:
    """
    Create invalid tours.

    :param random: the random number generator to use
    :param y: the input tour
    :returns: the invalid version of `y`
    """
    ly: Final[int] = len(y)
    y[0:ly] = range(ly)
    random.shuffle(y)
    yorig = np.copy(y)

    ri = random.integers
    while np.all(y == yorig):
        if ri(2) <= 0:
            z1 = z2 = ri(ly)
            while z1 == z2:
                z2 = ri(ly)
            y[z1] = y[z2]
        if ri(2) <= 0:
            y[ri(ly)] = ri(ly, 10 * ly)
        if ri(2) <= 0:
            y[ri(ly)] = ri(-2 * ly, -1)
    return y


def validate_algorithm_on_1_tsp(
        algorithm: Algorithm | Callable[
            [Instance, Permutations], Algorithm],
        instance: str | None = None, max_fes: int = 256,
        random: Generator = __RANDOM) -> None:
    """
    Check the validity of a black-box algorithm on one TSP instance.

    :param algorithm: the algorithm or algorithm factory
    :param instance: the instance name, or `None` to randomly pick one
    :param max_fes: the maximum number of FEs
    :param random: the default random generator to use
    """
    if not (isinstance(algorithm, Algorithm) or callable(algorithm)):
        raise type_error(algorithm, "algorithm", Algorithm, True)
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "TSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded bin TSP instance {instance!r}",
                         Instance)

    search_space = Permutations.standard(inst.n_cities)
    objective = TourLength(inst)
    if callable(algorithm):
        algorithm = algorithm(inst, search_space)
    if not isinstance(algorithm, Algorithm):
        raise type_error(algorithm, "algorithm", Algorithm, call=True)

    validate_algorithm(algorithm=algorithm,
                       solution_space=search_space,
                       objective=objective,
                       max_fes=max_fes)


def validate_algorithm_on_tsp(
        algorithm: Callable[[Instance, Permutations], Algorithm],
        symmetric: bool = True, asymmetric: bool = True,
        max_fes: int = 256, random: Generator = __RANDOM) -> None:
    """
    Validate an algorithm on a set of TSP instances.

    :param algorithm: the algorithm factory
    :param symmetric: include symmetric instances
    :param asymmetric: include asymmetric instances
    :param max_fes: the maximum FEs
    :param random: the random number generator
    """
    for i in tsp_instances_for_tests(random, symmetric, asymmetric):
        validate_algorithm_on_1_tsp(algorithm, i, max_fes, random)


def validate_objective_on_1_tsp(
        objective: Objective | Callable[[Instance], Objective],
        instance: str | None = None,
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on 1 TSP instance.

    :param objective: the objective function or a factory creating it
    :param instance: the instance name
    :param random: the random number generator
    """
    if instance is None:
        instance = str(random.choice(Instance.list_resources()))
    if not isinstance(instance, str):
        raise type_error(instance, "TSP instance name", (str, None))
    inst = Instance.from_resource(instance)
    if not isinstance(inst, Instance):
        raise type_error(inst, f"loaded TSP instance {instance!r}",
                         Instance)

    if callable(objective):
        objective = objective(inst)

    validate_objective(
        objective=objective,
        solution_space=Permutations.standard(inst.n_cities),
        make_solution_space_element_valid=make_tour_valid,
        is_deterministic=True)


def validate_objective_on_tsp(
        objective: Objective | Callable[[Instance], Objective],
        symmetric: bool = True, asymmetric: bool = True,
        random: Generator = __RANDOM) -> None:
    """
    Validate an objective function on TSP instances.

    :param objective: the objective function or a factory creating it
    :param symmetric: include symmetric instances
    :param asymmetric: include asymmetric instances
    :param random: the random number generator
    """
    for i in tsp_instances_for_tests(random, symmetric, asymmetric):
        validate_objective_on_1_tsp(objective, i, random)



"""
Several possible controllers for the dynamic control scenarios.

- :mod:`~moptipyapps.dynamic_control.controllers.linear` offers simple linear
  controllers.
- :mod:`~moptipyapps.dynamic_control.controllers.quadratic` offers simple
  quadratic controllers.
- :mod:`~moptipyapps.dynamic_control.controllers.cubic` offers simple
  cubic controllers.
- :mod:`~moptipyapps.dynamic_control.controllers.ann` offers poor man's
  artificial neural networks (ANNs), i.e., networks represented as plain
  functions whose parameter vectors are subject to optimization.
- :mod:`~moptipyapps.dynamic_control.controllers.predefined` provides a set of
  controllers taken from the works of NOACK, CORNEJO MACEDA, LI, and SUN.
- :mod:`~moptipyapps.dynamic_control.controllers.partially_linear` offers
  several linear controllers anchored at certain points in the state space and
  always uses the controller closest to the current state.
- :mod:`~moptipyapps.dynamic_control.controllers.min_ann` offers controllers
  similar to :mod:`~moptipyapps.dynamic_control.controllers.min_ann`, but
  instead of using the ANN output as controller value, these ANNs have an
  additional input `z` and then use the value `z*` as controller output for
  which the ANNs take on the smallest value.
- :mod:`~moptipyapps.dynamic_control.controllers.peaks` provides controllers
  similar to ANNs but using peak functions (`exp(-a²)`) as activation
  functions.

These controller blueprints can be used in two ways. The basic use case is to
synthesize, well, controllers for the dynamic system. In the dynamic system
controller synthesis problem, we have a system of differential equations
`D=ds/dt` where `D` depends on the current system state `s` and the output of
a controller `c`. So `D` is actually a function `D(s, c)`. Now the goal is to
get the dynamic system to move into a nice and stable state. We want to find
a controller `c` that can do this. Now `c` is actually a function `c(s, p)`
that takes as input the current system state `s` and its own parameterization
`p`. Imagine `c` to be, for example, an artificial neural network (an
:mod:`~moptipyapps.dynamic_control.controllers.ann`), then `p` is its weight
vector and `p` is subject to optimization. We would try to find the values `p`
that minimize a certain :mod:`~moptipyapps.dynamic_control.objective` function
that could, let's say, represent the cost or energy in the system. So use case
number one of our controllers is to represent the controller blueprints for
this process.

Use case number two is to use the controllers as blueprints for system models
`M`. Basically, a system model `M` should be a drop-in replacement for the
system equations `D`. In the *real world*, obviously, we do not have the
system equations that govern a complex system like a helicopter or something.
But we can test a helicopter controller in a wind tunnel. In such a scenario,
evaluating the objective function is very costly. If, instead, we could learn
a model `M` that can reasonably accurately describe the helicopter-experiment
behavior `D`, then we could replace the actual experiment with a simulation.
Since computing power is cheap, that would be awesome. Now it turns out that
the structure of the equations, i.e., a parameterized function relating inputs
to outputs, that we need for this, is pretty much the same as in use case
number one above. So we can re-use the same code, the basic
:mod:`~moptipyapps.dynamic_control.controller` blueprints, also for the
purpose of synthesizing models using a
:mod:`~moptipyapps.dynamic_control.model_objective` function.
"""
